from .resource import ZERO

from math import inf as INFIN
from contextlib import contextmanager
from functools import partial
from inspect import signature

import trio

def open_waitgroup():
    if not CombineGroup:
        raise RuntimeError("`open_waitgroup` is only valid inside `grain.combine.run`")
    return CombineGroup()

async def exec1(res, fn, *args, **kwargs):
    if not Exec1:
        raise RuntimeError("`exec1` is only valid inside `grain.combine.run`")
    return await Exec1(res, fn, *args, **kwargs)

async def load_cache_or_exec1(res, fn, *args, **kwargs):
    """Run `fn.cache_fn` before running a job, then either
    return the cached value or submit the job.
    The cache function's signature should be a subset
    of `fn`'s signature, with order preserved.
    If `fn.cache_res` is not `ZERO`, the
    cache_fn will be submitted to the queue.
    """
    sig_c = list(signature(fn.cache_fn).parameters.keys()) # unlike `getfullargspec` this unwrap decorators
    kwargs_c = {}
    while len(sig_c) > 0 and sig_c[-1] in kwargs:
        k = sig_c.pop()
        kwargs_c[k] = kwargs[k]
    sig = list(signature(fn).parameters.keys())
    args_c, i = [], 0
    for k in sig_c:
        while i < len(args) and sig[i] != k: i+=1
        if i == len(args): raise ValueError(f"Cannot find a positional arg for parameter {k}")
        args_c.append(args[i])

    if fn.cache_res is ZERO: # "free" cache_fn
        try:
            ok, r = await fn.cache_fn(*args_c, **kwargs_c) # TODO: provide sync alternative?
        except Exception:
            ok = False
    else:
        ok, r = await exec1(fn.cache_res, fn.cache_fn, *args_c, **kwargs_c)
    if ok:
        return r
    return await exec1(res, fn, *args, **kwargs)

async def _grouped_task(gid, fn, *args, **kwargs):
    r = await fn(*args, **kwargs)
    return gid, r

async def _run_and_send_result(chan, fn, *args, **kwargs):
    r = await fn(*args, **kwargs)
    chan.send_nowait(r)


CombineGroup, Exec1 = None, None
@contextmanager
def CombineGroup_ctxt(exer, push_newgroup):
    sema = trio.Semaphore(300) # rate limit
    gid = 0 # monotonously incremental with __aenter__'s side-effect

    class __CombineGroup(object):
        """A wait group that submit jobs to backend
        GrainExecutor and collect the results.
        This context manager is reusable, but single-use
        is recommended, because creating a instance
        each time allows nesting multiple ones.
        """
        def __init__(self):
            self.counter = 0
            self.results = []
        async def asubmit(self, res, fn, *args, **kwargs):
            await exer.asubmit(res, _grouped_task, self.gid, fn, *args, **kwargs)
            self.counter += 1
        def submit(self, res, fn, *args, **kwargs):
            exer.submit(res, _grouped_task, self.gid, fn, *args, **kwargs)
            self.counter += 1
        def start_subtask(self, fn, *args, **kwargs):
            # tasks requesting zero resource would be executed locally
            s, r = trio.open_memory_channel(1)
            exer.submit(ZERO, _grouped_task, self.gid, r.receive)
            _gn.start_soon(partial(_run_and_send_result, s, fn, *args, **kwargs))
            self.counter += 1
        def load_cache_or_submit(self, res, fn, *args, **kwargs):
            self.start_subtask(load_cache_or_exec1, res, fn, *args, **kwargs)
        async def __aenter__(self): # async part of __init__
            if self.counter > 0:
                raise RuntimeError("attempt to re-enter a CombineGroup. For recursive use, create a new instance instead.")
            await sema.acquire()
            nonlocal gid
            self.gid = gid = gid + 1
            s, self.resultq = trio.open_memory_channel(INFIN)
            push_newgroup.send_nowait( (-1, (self.gid, s)) )
            return self
        async def __aexit__(self, *exc):
            if any(exc): return False
            resultd = {}
            async with self.resultq:
                if self.counter > 0:
                    async for i, r in self.resultq:
                        resultd[i] = r
                        self.counter -= 1
                        if self.counter == 0: break
            self.results = [v for k,v in sorted(resultd.items(), key=lambda x: x[0])]
            sema.release()

    async def __Exec1(res, fn, *args, **kwargs):
        nonlocal gid
        g = gid = gid + 1
        sq, rq = trio.open_memory_channel(1)
        push_newgroup.send_nowait( (-1, (g, sq)) )
        exer.submit(res, _grouped_task, g, fn, *args, **kwargs)
        async with rq, sq:
            return (await rq.receive())[1]

    global CombineGroup, Exec1
    CombineGroup, Exec1 = __CombineGroup, __Exec1
    try:
        yield
    finally:
        CombineGroup, Exec1 = None, None

async def relay(inq):
    # inq has two senders: push_newgroup; push_result.
    # This is an analogue of `select` in Golang
    outqs = {}
    async with inq:
        async for i, (g, outq_or_rslt) in inq:
            if i == -1:
                outqs[g] = outq_or_rslt # init outq
            else:
                try:
                    outqs[g].send_nowait((i, outq_or_rslt))
                except trio.BrokenResourceError:
                    # This usually happens when one of the main subtasks throws an error,
                    # cancelling receive channel's task.
                    # Suppress this so that the real error can surface.
                    break

async def boot_combine(subtasks, args, kwargs):
    from .head import GrainExecutor
    from .combine import relay, CombineGroup_ctxt # no global dependency
    from .config import load_conf
    from collections.abc import Iterable
    config = load_conf(kwargs.pop('config_file', None), 'head')
    async with trio.open_nursery() as _n, \
               GrainExecutor(_n=_n, config=config, *args, **kwargs) as exer, \
               exer.push_result.clone() as push_newgroup:
        _n.start_soon(relay, exer.resultq)
        global _gn # for all subtasks
        with CombineGroup_ctxt(exer, push_newgroup):
            async with trio.open_nursery() as _gn:
                if isinstance(subtasks, Iterable):
                    for st in subtasks:
                        _gn.start_soon(st)
                else:
                    await subtasks()

def run(subtasks, *args, **kwargs):
    trio.run(boot_combine, subtasks, args, kwargs)



import trio
from trio import socket

import struct
import io
from contextlib import contextmanager, asynccontextmanager
from math import inf as INFIN
from urllib.parse import urlsplit
import logging
logger = logging.getLogger(__name__)

RENDEZVOUS_TIMEOUT = 10
CONNECT_RETRY_INTERVAL = 0.5

# +--------+-----------+
# | LEN(4) | DATA(LEN) |
# +--------+-----------+
FMT, LEN = '>L', 4 # unsigned long / uint32


async def bound_socket(local_addr):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    if hasattr(socket, "SO_REUSEPORT"):
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    await s.bind(local_addr)
    logger.debug(f"socket bound to {local_addr} created")
    return s

async def accept(s, win_s):
    with closing_otherwise(s):
        logger.debug("Accepting...")
        s.listen()
        conn, _ = await s.accept()
        logger.debug("Accepted")
        win_s.send_nowait(conn)

async def connect(s, peer_addr, win_s):
    with closing_otherwise(s):
        logger.debug(f"Connecting to {peer_addr}")
        while True:
            try:
                await s.connect(peer_addr)
            except Exception as e:
                logger.debug(e)
                await trio.sleep(CONNECT_RETRY_INTERVAL)
                continue
            else:
                logger.debug(f"Connected {peer_addr}")
                break
        win_s.send_nowait(s)

async def init_conn(bridge, iface=""):
    *sarg, _, rbridge = (await socket.getaddrinfo(*bridge, type=socket.SOCK_STREAM))[0]
    retry = 3
    for _ in range(retry):
        sa = socket.socket(*sarg)
        sa.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        await sa.connect(rbridge)
        priv_host, priv_port = sa.getsockname() # self, private
        if iface:
            priv_host, priv_port = get_priv_host(iface), 0
        try:
            s_cona = await bound_socket((priv_host, priv_port)) # for connect, a
            if iface:
                _, priv_port = s_cona.getsockname()
            s_conb = await bound_socket((priv_host, priv_port)) # for connect, b
            s_acp = await bound_socket(('', priv_port)) # for accept
            break
        except OSError: # NOTE: albeit SO_REUSEADDR set, bind still occassionally fails
            logger.debug(f"cannot bind to private addr {make_ipaddr(priv_host, priv_port)}, retry")
            sa.close()
            await trio.sleep(0.1)
            continue
    else:
        raise OSError("failed to bind to a local address")
    priv_addr = make_ipaddr(priv_host, priv_port)
    logger.debug(f"self's private addr {priv_addr}")
    return trio.SocketStream(sa), priv_addr, s_acp, s_cona, s_conb

async def hole_punching(peer, s_acp, s_cona, s_conb):
    peer_publ_addr, peer_priv_addr = peer
    peer_publ_addr, peer_priv_addr = parse_ipaddr(peer_publ_addr), parse_ipaddr(peer_priv_addr)
    logger.debug(f"peer's public addr {peer_publ_addr}, private addr {peer_priv_addr}")

    win_s, win_r = trio.open_memory_channel(1)
    async with trio.open_nursery() as _n:
        _n.start_soon(accept, s_acp, win_s)
        _n.start_soon(connect, s_cona, peer_publ_addr, win_s)
        if peer_priv_addr != peer_publ_addr:
            _n.start_soon(connect, s_conb, peer_priv_addr, win_s)
        else:
            s_conb.close()
        conn = None
        with trio.move_on_after(RENDEZVOUS_TIMEOUT):
            conn = trio.SocketStream(await win_r.receive())
        if conn is None:
            raise OSError(f"failed to established a P2P connection with {peer}")
        _n.cancel_scope.cancel()
    return conn

"""2RTT per rendz
head            ===INIT,{key},===>           bridge
...
subd      ---SUBD,{key},{subd_priv}--->      bridge
head     <==={subd_publ}|{subd_priv}===      bridge
head   ---HEAD,{subd_publ},{head_priv}--->   bridge
head enters rendz
subd     <---{head_publ}|{head_priv}---      bridge
subd enters rendz
"""
async def open_tcp_stream_to_head(*, bridge, key, iface=""):
    """Open a P2P connection to head.
    """
    sa, priv_addr, s_acp, s_cona, s_conb = await init_conn(bridge, iface)
    async with sa:
        await send_packet(sa, f"SUBD,{key},{priv_addr}".encode())
        head = (await recv_packet(sa)).decode().split('|')
    return await hole_punching(head, s_acp, s_cona, s_conb)

async def serve_tcp_p2p(handler, *, bridge, key, iface="", handler_nursery=None, task_status=trio.TASK_STATUS_IGNORED):
    """Subscribe to subds' connection requests.
    For each request, open a P2P connection to the subd.
    """
    async with await trio.open_tcp_stream(*bridge, happy_eyeballs_delay=INFIN) as s, \
               use_or_open_nursery(handler_nursery) as _n:
        await send_packet(s, f"INIT,{key},".encode())
        task_status.started()
        async for subd in iter_packet(s):
            subd_publ_addr, subd_priv_addr = subd.decode().split('|')
            with suppress_then_log(OSError):
                sa, priv_addr, s_acp, s_cona, s_conb = await init_conn(bridge, iface)
                async with sa:
                    await send_packet(sa, f"HEAD,{subd_publ_addr},{priv_addr}".encode())
                stream = await hole_punching((subd_publ_addr, subd_priv_addr), s_acp, s_cona, s_conb)
                _n.start_soon(_run_handler, stream, handler)
        raise RuntimeError("bridge quits unexpectedly")


def get_priv_host(iface):
    import psutil
    i = psutil.net_if_addrs()[iface]
    return next(x.address for x in i if x.family == socket.AF_INET)

def parse_ipaddr(host_port):
    r = urlsplit('//' + host_port)
    return r.hostname or "0.0.0.0", r.port

def make_ipaddr(host, port):
    return f"[{host}]:{port}" if ':' in host else f"{host}:{port}"

async def send_packet(s: trio.SocketStream, data):
    await s.send_all(struct.pack(FMT, len(data)) + data)

async def recv_packet(s: trio.abc.ReceiveStream):
    plen, = struct.unpack(FMT, await receive_exactly(s, LEN))
    return await receive_exactly(s, plen)

async def iter_packet(s: trio.abc.ReceiveStream):
    while True:
        try:
            yield await recv_packet(s)
        except (EOFError, trio.BrokenResourceError):
            return

async def receive_exactly(s: trio.abc.ReceiveStream, n):
    buf = io.BytesIO()
    while n > 0:
        # NOTE: Usually when the other end quits (gracefully
        # or not), `s.receive_some` returns b"" (indicating EOF).
        # However, this is not always the case for a P2P connection.
        data = await s.receive_some(n)
        if data == b"":
            raise EOFError("end of receive stream")
        buf.write(data)
        n -= len(data)
    return buf.getvalue()

@contextmanager
def closing_otherwise(s):
    try:
        yield
    except BaseException:
        s.close()
        raise

@asynccontextmanager
async def use_or_open_nursery(_n):
    if _n is not None:
        yield _n
    else:
        async with trio.open_nursery() as _n:
            yield _n

async def _run_handler(s, handler):
    try:
        await handler(s)
    finally:
        await trio.aclose_forcefully(s)

@contextmanager
def suppress_then_log(exctype=Exception):
    try:
        yield
    except exctype as e:
        logger.warning(f"{e.__class__.__name__}: {e}")



from .pair import *

import pytest
import trio
trio_open_tcp_listeners = trio.open_tcp_listeners
from trio.testing import MockClock, open_stream_to_socket_listener

from functools import wraps

def mock_network(fn):
    listeners = {}
    async def open_tcp_listeners(port, *, host, **__):
        l, *_ = await trio_open_tcp_listeners(0) # take only one
        nonlocal listeners
        listeners[(host,port)] = l
        return [l]
    async def open_tcp_stream(host, port, **__):
        return await open_stream_to_socket_listener(listeners[(host,port)])

    @wraps(fn)
    async def __wrapped(monkeypatch):
        monkeypatch.setattr(trio, "open_tcp_listeners", open_tcp_listeners)
        monkeypatch.setattr(trio, "open_tcp_stream", open_tcp_stream)
        return await fn(monkeypatch)
    return __wrapped


@mock_network
async def test_signal(monkeypatch):
    # plain
    async def _listen(_n, task_status=trio.TASK_STATUS_IGNORED):
        async with listen_signal("tcp://test:8000", _n) as l:
            task_status.started()
            async for (_, m) in l:
                assert m == b"msg"
                break
            else:
                assert False, "listen_signal close before receiving"
    async with trio.open_nursery() as _n:
        await _n.start(_listen, _n)
        await notify("tcp://test:8000", b"msg")

    # segment
    async def _listen_sc(_n, task_status=trio.TASK_STATUS_IGNORED):
        async with SocketChannel("tcp://test:8001", _n=_n, listen=True) as l:
            task_status.started()
            async for m in l:
                assert m == dict(cmd="test")
                break
            else:
                assert False, "SocketChannel close before receiving"
    async with trio.open_nursery() as _n:
        await _n.start(_listen_sc, _n)
        await notify("tcp://test:8001", dict(cmd="test"), seg=True)



from contextvars import ContextVar

_GVAR = dict(
    res      = ContextVar("GRAIN_RESOURCE"),
    instance = ContextVar("GRAIN_INSTANCE"),
    context  = ContextVar("GRAIN_CONTEXT"),
)

_None = object()

class GrainVar:

    def __getattr__(self, k):
        return _GVAR.get(k).get()
    def __setattr__(self, k, v):
        _GVAR.get(k).set(v)

    def __getstate__(self):
        return { k:v.get() for k,v in _GVAR.items() if v.get(_None) is not _None }
    def __setstate__(self, state):
        for k,v in state.items(): setattr(self, k, v)

GVAR = GrainVar()



"""Packet handling with msgpack, compatible with functions from `conn.py`.
This file will eventually merge into `conn.py` once we finish migrating all
communications to msgpack.
"""
import msgpack
import trio

from . import resource

_p = msgpack.Packer()

async def send_packet(s: trio.SocketStream, obj):
    if 'res' in obj:
        obj['res'] = obj['res'].encode_msgp()
    await s.send_all(_p.pack(obj))

def decode_res(obj):
    if 'res' in obj:
        res = resource.ZERO
        for rn, rargs in obj['res'].items():
            res &= getattr(resource, rn)(**rargs)
        obj['res'] = res
    return obj

async def iter_packet(s: trio.abc.ReceiveStream):
    _u = msgpack.Unpacker()
    while True:
        try:
            data = await s.receive_some()
        except trio.BrokenResourceError:
            return
        if data == b"": # end of receive stream
            return
        _u.feed(data)
        for msg in _u:
            if type(msg) is not dict:
                print("received malformed packet:", data)
                return
            yield decode_res(msg)

async def recv_packet(s: trio.abc.ReceiveStream):
    try:
        return await (iter_packet(s).__aiter__().__anext__())
    except StopAsyncIteration:
        raise EOFError("end of receive stream")



"""Edge protocol: connection discovery depending on network filesystem only
"""
import trio
from trio import socket
import psutil

from pathlib import Path
import json
import os
import fcntl
import logging
logger = logging.getLogger(__name__)

from .conn import use_or_open_nursery, _run_handler, make_ipaddr

HOST = socket.gethostname()

# """edge file spec"""
# class Listener:
#     host: str
#     lid:  int
#     urls: dict[str, str] # iface: url
# class Dialer:
#     host: str
#     urls: dict[str, str] # iface: url
# class EdgeFile:
#     listener: Listener
#     dialer:   list[Dialer]

async def open_tcp_edge_stream(*, edge_file):
    edge_file = Path(edge_file)
    edge_lock = edge_file.with_suffix('.lock')
    async with UnixFileLock(edge_lock):
        info = {}
        if edge_file.exists():
            info = json.loads(edge_file.read_text())
        if (linfo := info.get('listener')) is not None: # try connecting to the listener
            logger.debug(f"Try connecting to listener {linfo['host']}")
            urls = linfo['urls']
            if linfo['host'] == HOST and 'UDS' in urls: # use Unix domain socket if local
                urls = dict(UDS=urls['UDS'])
            stream = await try_urls(urls.items())
            if stream is not None:
                return stream
            logger.warning(f"Failed to established a connection with {linfo['host']}")

        # fall back: listen & enqueue self
        listeners, urls = await listen_all_iface()
        info.setdefault('dialer', []).append(dict(
            host=HOST, urls=urls,
        ))
        edge_file.write_text(json.dumps(info))
    stream = None
    async def _one_listener(l, cs):
        nonlocal stream
        stream = await l.accept()
        cs.cancel()
    async with trio.open_nursery() as _n:
        for l in listeners:
            _n.start_soon(_one_listener, l, _n.cancel_scope)
    return stream

async def serve_tcp_edge(handler, *, edge_file, handler_nursery=None, task_status=trio.TASK_STATUS_IGNORED):
    """Try connecting to all pending dialers, then listen and
    annonce itself.
    """
    edge_file = Path(edge_file)
    edge_lock = edge_file.with_suffix('.lock')
    async with use_or_open_nursery(handler_nursery) as _n:
        started = False
        async with UnixFileLock(edge_lock):
            if edge_file.exists():
                info = json.loads(edge_file.read_text())
                if (linfo := info.get('listener', None)) is not None:
                    logger.warning(f"Using an edge file claimed by another listener on host {linfo['host']}, "
                                    "which might be running or didn't exit cleanly.")
                for dinfo in info.get('dialer', []): # try connecting to all pending dialers
                    logger.debug(f"Try connecting to pending dialer {dinfo['host']}")
                    stream = await try_urls(dinfo['urls'].items())
                    if stream is None:
                        logger.warning(f"Failed to establish a connection with {dinfo['host']}")
                        continue
                    _n.start_soon(_run_handler, stream, handler)
                    if not started: # eagerly admit if any handler might start running
                        task_status.started()
                        started = True

            # listen on every iface
            listeners, urls = await listen_all_iface()
            lid = hash(tuple(urls.values()))
            # override, even if a server has been started before the current one
            edge_file.write_text(json.dumps(dict(
                listener=dict(host=HOST, lid=lid, urls=urls),
            )))

        try:
            # TODO: Unix domain socket server
            await trio.serve_listeners(handler, listeners, handler_nursery=_n, task_status=task_status if not started else trio.TASK_STATUS_IGNORED)
        finally:
            with trio.move_on_after(1) as cleanup_scope:
                cleanup_scope.shield = True
                async with UnixFileLock(edge_lock):
                    if edge_file.exists() and json.loads(edge_file.read_text())['listener']['lid'] == lid:
                        edge_file.unlink() # delete only if no other server is running


async def try_urls(urls, happy_eyeballs_delay=0.250):
    from .pair import parse_url, open_tcp_stream # avoid cyclic import

    winning_stream = None
    async def attempt_connect(attempt_failed, *sargs):
        nonlocal winning_stream
        try:
            winning_stream = await open_tcp_stream(*sargs)
            _n.cancel_scope.cancel()
        except OSError as e:
            logger.debug(f"OSError: {e}")
            attempt_failed.set()

    async with trio.open_nursery() as _n:
        for ifn, url in urls:
            logger.debug(f"Trying {ifn}: {url}")
            proto, host, port, opts = parse_url(url)
            attempt_failed = trio.Event()
            _n.start_soon(attempt_connect, attempt_failed, proto, host, port, opts)
            with trio.move_on_after(happy_eyeballs_delay):
                await attempt_failed.wait()
    return winning_stream

async def listen_all_iface():
    listeners = {
        ifn: (await trio.open_tcp_listeners(0, host=ifaddr.address))[0]
        for ifn, ifaddrs in psutil.net_if_addrs().items() if ifn != 'lo'
            for ifaddr in ifaddrs if ifaddr.family in {socket.AF_INET}#, socket.AF_INET6}
    } # FIXME: ipv6
    urls = {
        ifn: f"tcp://{make_ipaddr(*s.socket.getsockname())}"
        for ifn, s in listeners.items()
    }
    return list(listeners.values()), urls

class UnixFileLock:
    POLL_TIME = 0.01
    def __init__(self, filename):
        self._filename = filename
        self._fd = None
    async def acquire(self):
        fd = os.open(self._filename, os.O_RDWR | os.O_CREAT | os.O_TRUNC)
        while True:
            try:
                fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except BlockingIOError:
                await trio.sleep(self.POLL_TIME)
            else:
                self._fd = fd
                return self
    async def release(self, *_):
        if self._fd is not None:
            fd, self._fd = self._fd, None
            fcntl.flock(fd, fcntl.LOCK_UN)
            os.close(fd)
    __aenter__ = acquire
    __aexit__ = release



try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib
from attrs import define, Factory
import cattr

from typing import Optional, List, Dict, Any, Literal
from os import environ as ENV
from collections import ChainMap
from io import StringIO
from pathlib import Path
import logging
logger = logging.getLogger(__name__)

@define(slots=False)
class Script:
    cores:         int = 0
    memory:        int = 0
    setup_cleanup: str = ""
    shebang:       str = "#!/bin/bash"
    queue:         str = "''"
    walltime:      str = ""
    extra_args:    List[str] = Factory(list)

@define
class CustomSystem:
    submit_cmd: str
    directory:  str
    template:   str

@define
class Config:
    system:        str = ""
    script:        Script = Factory(Script)
    contextmod:    str = ""
    custom_system: Optional[CustomSystem] = None
    address:       str = ""
    def __attrs_post_init__(self):
        if not self.address:
            (_local_share := Path.home()/".local/share").mkdir(mode=0o755, parents=True, exist_ok=True)
            self.address = f"edge://{_local_share}/edge-file-default"

@define
class Gnaw:
    enabled:    bool = True
    log_file:   str = ""
    max_conn:   int = 8
    idle_quit:  str = "30m"
    swarm:      int = 0
    extra_args: List[str] = Factory(list)

@define
class Head(Config):
    name:          str = "grain_head"
    main_log_file: str = "/dev/null"
    log_file:      str = ""
    listen:        str = ""
    cmd:           str = ""
    gnaw:          Gnaw = Factory(Gnaw)
    def __attrs_post_init__(self):
        Config.__attrs_post_init__(self)
        if not self.listen:
            self.listen = self.address

@define
class Worker(Config):
    specialized_type: str = ""
    name:             str = "w{{HHMMSS}}"
    log_file:         str = "w{{HHMMSS}}.log"
    dial:             str = ""
    cli_dial:         str = ""
    res:              Dict[str, Any] = Factory(dict)
    def __attrs_post_init__(self):
        Config.__attrs_post_init__(self)
        if not self.dial:
            self.dial = self.address
        if not self.cli_dial:
            self.cli_dial = self.dial

@define
class GenericConfig:
    head:   Head
    worker: Worker

_loose_filler = dict(
    head=dict(gnaw=dict(enabled=False)), # disable gnaw for test purpose
)

def ChainMapNested(d0, d1):
    for k in set(d0.keys()) & set(d1.keys()):
        if type(d0[k]) is dict and type(d1[k]) is dict:
            d0[k] = ChainMapNested(d0[k], d1[k])
    return ChainMap(d0, d1)

def load_conf(config=None, mode: Literal['', 'head', 'worker']=''):
    if config is False:
        logger.info("Config file is disabled, using default settings.")
        conf = _loose_filler
    else:
        config = config or ENV.get("GRAIN_CONFIG", "grain.toml")
        try:
            config_s = Path(config).read_bytes().decode() if type(config) is str else config.read()
            conf = tomllib.loads(config_s)
        except FileNotFoundError:
            logger.error(f"Cannot find Grain config file {config!r}")
            exit(1)
        if type(config) is StringIO: # internal config fragment
            conf = ChainMap(conf, _loose_filler)
    conf.setdefault('head', {})
    conf.setdefault('worker', {})
    if mode == 'worker':
        return cattr.structure(ChainMapNested(conf['worker'], conf), Worker)
    elif mode == 'head':
        return cattr.structure(ChainMapNested(conf['head'], conf), Head)
    return GenericConfig(
        worker=cattr.structure(ChainMapNested(conf['worker'], conf), Worker),
        head=cattr.structure(ChainMapNested(conf['head'], conf), Head),
    )

def load_conf_sworker(config=None):
    conf = load_conf(config, 'worker')
    if conf.specialized_type:
        return [conf]
    return []



from .worker import *
from .head import GrainExecutor
from .config import load_conf
from .resource import Memory, WTime
from .contextvar import GVAR

import pytest
import trio

from functools import partial

GrainExecutor = partial(GrainExecutor, config=load_conf(False, 'head'))

async def addone(x):
    return x+1

async def test_worker():
    N = 3
    async with trio.open_nursery() as _n, \
               GrainExecutor(_n=_n, nolocal=True) as exer:
        GVAR.instance = "test-worker-A"
        _n.start_soon(grain_worker, Memory(4)&WTime(T=10,countdown=True), exer.mgr.listen_addr)
        for i in range(N):
            exer.submit(Memory(2), addone, i)
    results = [None] * N
    async with exer.resultq:
        async for i, r in exer.resultq:
            results[i-1] = r
    assert results == list(range(1,N+1))



import trio
import dill as pickle

from functools import wraps, partial
from inspect import iscoroutinefunction
import sys
import os
from math import inf as INFIN
from subprocess import SubprocessError
from contextlib import asynccontextmanager

from grain.contextvar import GVAR # process local

IDLE_TIMEOUT = 60 # keep idle subprocess for 1min

def subprocify(fn):
    """Subprocify is a decorator that turns a sync function (usually a CPU-bound
    job) into an async job to be executed on an external Python interpreter
    subprocess. Please allocate at least 1 processor and adequate vmem for this job.
    """
    if iscoroutinefunction(getattr(fn, "func", fn)):
        raise TypeError("subprocify only wraps sync functions, but an async function is given")
    @wraps(fn)
    async def _wrapped(*args, **kwargs):
        if idle_subp:
            p = idle_subp.pop()
        else:
            p = await start_worker_process()
        ok, r = await p.execf(GVAR, partial(fn, *args, **kwargs))
        # FIXME: for now, if the job is cancelled from outside (e.g. timeout), the worker
        # process is abandoned, because we cannot determine the worker's status and output.
        if ok or type(r) is not BrokenSubprocessError:
            idle_subp.add(p)
        if not ok:
            raise r.with_traceback(r.__traceback__)
        return r
    return _wrapped

_sn = None
idle_subp = set()

@asynccontextmanager
async def subprocess_pool_scope(): # process local
    """Initialize the subprocess pool.
    This is required before any subprocified function is called.
    After the first call, this function is a no-op.
    """
    global _sn
    if _sn is not None and not _sn._closed:
        yield
        return
    async with trio.open_nursery() as _sn:
        yield
        _sn.cancel_scope.cancel()
async def subprocess_pool_daemon(task_status=trio.TASK_STATUS_IGNORED): # process local
    global _sn
    if _sn is not None and not _sn._closed:
        task_status.started(_sn.cancel_scope)
        return
    async with trio.open_nursery() as _sn:
        task_status.started(_sn.cancel_scope)
        await trio.sleep_forever()

async def start_worker_process():
    p = _Process()
    if _sn is None:
        raise RuntimeError("You need to initialize the subprocess pool by `subprocess_pool_scope`.")
    await _sn.start(p.loop)
    return p

class _Process(object):
    def __init__(self):
        self.to_child, self.from_child = None, None
        self.idle_timeout = trio.move_on_after(INFIN)
    async def loop(self, task_status=trio.TASK_STATUS_IGNORED):
        child_r, parent_w = os.pipe()
        parent_r, child_w = os.pipe()
        async with trio.lowlevel.FdStream(parent_w) as self.to_child, \
                   trio.lowlevel.FdStream(parent_r) as self.from_child:
            task_status.started()
            with self.idle_timeout:
                await trio.run_process([sys.executable, "-m", "grain.subproc_worker", "--fd-read", str(child_r), "--fd-write", str(child_w)],
                                       check=False,
                                       pass_fds=(child_r, child_w))
            # idle timeout: SIGINT, then remove self from the pool
            try:
                idle_subp.remove(self)
            except KeyError:
                pass
    async def execf(self, gvar, fn):
        self.idle_timeout.deadline = INFIN
        try:
            await pickle_dump(self.to_child, (gvar, fn))
            ok, r = await pickle_load(self.from_child)
            self.idle_timeout.deadline = trio.current_time() + IDLE_TIMEOUT
            return ok, r
        except trio.ClosedResourceError as e: # pipe broken / subprocess die
            self.idle_timeout.cancel() # termiate process immediately
            return False, BrokenSubprocessError("broken pipe to subprocess").from_(e)
    async def aclose(self): # NOTE: not used
        idle_subp.remove(self)
        await pickle_dump(self.to_child, None) # FIN

async def pickle_dump(f, obj):
    d = pickle.dumps(obj)
    await f.send_all(d)
async def pickle_load(f):
    d = await f.receive_some()
    while d[-1] != 46: # '.'; legit b/c the worker will not start another job until we drain the result
        d += await f.receive_some()
    return pickle.loads(d)

class BrokenSubprocessError(SubprocessError):
    """Exception raised when a subprocess crashes due to an unrecoverable error.
    """
    def from_(self, e):
        self.__cause__, self.__traceback__ = e, e.__traceback__
        return self



from ._version import __version__
from .contextvar import GVAR

import logging
logging.basicConfig(format='[%(asctime)s] %(name)s:%(levelname)s: %(message)s')
logging.getLogger('grain').setLevel(logging.INFO)



from .head import *
from .resource import Memory, ZERO
from .config import load_conf

import pytest
import trio
from trio.testing import MockClock
try:
    from exceptiongroup import ExceptionGroup  # for Python < 3.11
except ImportError:
    pass

from functools import partial
from io import StringIO

sprint = partial(trio.run, clock=MockClock(rate=1000))
GrainExecutor = partial(GrainExecutor, config=load_conf(False, 'head'))

async def rev(n, i):
    await trio.sleep(n-i)
    return i

def test_order(monkeypatch):
    monkeypatch.setattr(trio, "run", sprint)
    N = 10
    with GrainExecutor([], Memory(8)) as exer:
        for i in range(N):
            exer.submit(Memory(2), rev, N, i)
    assert exer.results == list(range(N))


class Critical(Exception):
    pass
class Temporary(Exception):
    pass

async def critical_err_task():
    raise Critical
trial = 0
async def temporary_err_task():
    global trial
    trial += 1
    raise Temporary

def test_local_critical():
    with pytest.raises(ExceptionGroup) as excinfo, \
         GrainExecutor() as exer:
        exer.submit(ZERO, critical_err_task)
    assert excinfo.group_contains(RuntimeError, match="local worker quit")

def test_local_lastjob_retry():
    with pytest.raises(ExceptionGroup) as excinfo, \
         GrainExecutor(temporary_err=(Temporary,)) as exer:
        exer.submit(ZERO, temporary_err_task)
    assert excinfo.group_contains(RuntimeError, match="local worker quit")
    assert trial == FULL_HEALTH


async def say_something():
    print("something")

def test_local_redirectouterr(capsys):
    with GrainExecutor() as exer:
        exer.submit(ZERO, say_something)
    captured = capsys.readouterr()
    assert "something\n" in captured.out

    with GrainExecutor(config=load_conf(
        StringIO('[head]\nlisten = "tcp://:4243"\nlog_file = "/dev/null"'), 'head'
    )) as exer:
        exer.submit(ZERO, say_something)
    captured = capsys.readouterr()
    assert "something\n" not in captured.out


class MockRemote(GrainPseudoRemote):
    def __init__(self, name):
        super().__init__(name, Memory(1))
    async def connect(self, _n):
        pass

async def test_manager(autojump_clock):
    def _assert_qlen(q, l):
        assert len(q._state.data) == l
    def _from_result_take_exactly(n):
        _assert_qlen(exer.resultq, n)
        [exer.resultq.receive_nowait() for _ in range(n)]
    async def _job():
        await trio.sleep(3)
        return "retval"

    pool = map(MockRemote, ["w0", "u0", "w1"])
    async with trio.open_nursery() as _n, \
               GrainExecutor(_n=_n) as exer:
        # release queued job when there're new resources
        exer.submit(Memory(1), _job)
        for w in pool:
            await exer.mgr.register(w, _n)
        await trio.sleep(3.1)
        _from_result_take_exactly(1)

        # terminate waits for pending jobs to finish
        exer.submit(Memory(1), _job)
        exer.submit(Memory(1), _job)
        _assert_qlen(exer.resultq, 0)
        await trio.sleep(.1)
        _assert_qlen(exer.push_job, 0)
        await exer.mgr.terminate("w?")
        assert len(exer.mgr.pool) == 1+3
        await trio.sleep(3)
        _from_result_take_exactly(2)
        assert len(exer.mgr.pool) == 1+1 and exer.mgr.pool[1].name == "u0"

        # unregister cancels all pending jobs
        exer.submit(Memory(1), _job)
        await trio.sleep(.1)
        _assert_qlen(exer.resultq, 0)
        await exer.mgr.unregister("*")
        assert len(exer.mgr.pool) == 1+0
        _assert_qlen(exer.resultq, 0)
        # cleanup
        await exer.mgr.register(MockRemote("_cleanup"), _n)
        await trio.sleep(3.1)
        _from_result_take_exactly(1)



import dill as pickle
import psutil
from trio import TooSlowError

import os
import signal
import argparse
from math import isinf

def exerf(GVAR, func):
    globals()['GVAR'] = GVAR # for func's convenience
    timeout = getattr(GVAR.res, 'T', 0)
    if isinf(timeout):
        timeout = 0
    if timeout:
        signal.signal(signal.SIGALRM, raise_timeout)
        signal.alarm(timeout)
    core = getattr(GVAR.res, 'c', [])
    try:
        if len(core) == 0: # NOTE: This Python interpreter proc by default takes 1 proc and ~40mb vmem.
            raise TypeError(f"Subprocess job resource expects at least core, get {GVAR.res} instead")
        psutil.Process().cpu_affinity(core)
        return True, func()
    except BaseException as e:
        if type(e) is KeyboardInterrupt:
            raise
        return False, e
    finally:
        if timeout:
            signal.signal(signal.SIGALRM, signal.SIG_IGN)

def raise_timeout(signum, frame):
    raise TooSlowError()


if __name__ == "__main__":
    argp = argparse.ArgumentParser(description="Subprocess worker that execute sync functions")
    argp.add_argument('--fd-read', type=int)
    argp.add_argument('--fd-write', type=int)
    A = argp.parse_args()

    with os.fdopen(A.fd_read, "rb", closefd=True) as from_parent, \
         os.fdopen(A.fd_write, "wb", closefd=True) as to_parent:
        while True:
            msg = pickle.load(from_parent)
            if msg is None: # FIN
                break
            r = exerf(*msg)
            pickle.dump(r, to_parent)
            to_parent.flush()



from .resource import Reject

import time
from datetime import datetime
from bisect import bisect_right
from contextlib import asynccontextmanager
import logging
logger = logging.getLogger(__name__)

import trio

MAXSPAN = 30 # in min

punch_event = {}

def log_event(ev):
    now = time.time()
    l = punch_event.setdefault(ev, [])
    l.append(now)
    if len(l) > 1000: roll(l, now)

time_event = {}

def log_timestart(ev, uid):
    tstmp, span, pending = time_event.setdefault(ev, ([], [], {}))
    pending[uid] = time.time()
def log_timeend(ev, uid):
    now = time.time()
    tstmp, span, pending = time_event[ev]
    tstmp.append(now)
    span.append(now-pending[uid])
    del pending[uid]
    if len(tstmp) > 1000:
        roll(tstmp, now)
        del span[:len(span)-len(tstmp)]

def tally(span):
    if span > MAXSPAN:
        raise ValueError(f"span {span} is larger than MAXSPAN {MAXSPAN}")
    now = time.time()
    r = {}
    for k,l in punch_event.items():
        roll(l, now)
        count = len(l) - bisect_right(l, now-span*60)
        if count: r[k] = count
    for k,(ts,sp,_) in time_event.items():
        i0 = bisect_right(ts, now-span*60)
        frq = len(ts) - i0
        if frq: r[k] = f"{sum(sp[i0:])/frq:.3f}s*{frq}"
    return r

@asynccontextmanager
async def stat_logger(span):
    async def _loop():
        while True:
            await trio.sleep(span*60)
            s = tally(span)
            if s:
                logger.info(s)
    async with trio.open_nursery() as _n:
        _n.start_soon(_loop)
        try:
            yield
        finally:
            _n.cancel_scope.cancel()

def roll(l, now=None):
    del l[:bisect_right(l, (now or time.time())-MAXSPAN*60)]


RES_TYPES = ("Cores", "Memory")
def ls_worker_res(workers):
    fs = ""
    len_name = max(len(w.name) for w in workers) + 4
    FMT = "{:>"+str(len_name)+"}" + "{:>10}"*len(RES_TYPES) + "  {}\n"
    def row(*cel):
        nonlocal fs
        fs += FMT.format(*cel)
    def ratio_map(pqs):
        return (f"{p}/{q}" for p,q in pqs)
    total = [[0,0] for _ in range(len(RES_TYPES))]
    row("", *RES_TYPES, "")
    for w in workers:
        wstat = w.res.stat()
        one = []
        for i,rt in enumerate(RES_TYPES):
            p,q = wstat.get(rt, (0,0))
            total[i][0] += p; total[i][1] += q
            one.append([p,q])
        row(w.name, *ratio_map(one), "paused" if type(w.res) is Reject else "")
    row("TOTAL", *ratio_map(total), "")
    return fs


PROBE_FN = {}
def reg_probe(ev, fn):
    PROBE_FN[ev] = fn

def collect_probe():
    ent = []
    for ev, fn in PROBE_FN.items():
        ent.append(f"{ev}: {fn()}")
    return '\n'.join(ent)



import click
import trio

from .pair import SocketChannel, notify
from .config import load_conf
from .util import load_mod
from ._version import __version__

from datetime import datetime
import subprocess
import tempfile
from time import sleep
import functools
from contextlib import contextmanager
import sys
import os
import json

VAR = { # replace any occurance of `{{key}}` with `value()`. Evaluate at each submission
    "HHMMSS": (lambda: datetime.now().strftime('%H%M%S')),
}

WORKER_MAIN = """echo "worker {c.name}'s node is $(hostname)"

{c.script.setup}

date
python -m grain.worker --url {c.dial!r} --res {c.script.res_str!r} --context {c.contextmod!r} --sworker {c.specialized_type!r}
date

{c.script.cleanup}
"""
HEAD_MAIN = """echo "head's node is $(hostname)"

{c.script.setup}

{c.cmd}

{c.script.cleanup}
"""

DFLT_TEMP = """{c.script.shebang}

{{MAIN}}
"""

SLURM_TEMP = """{c.script.shebang}
#SBATCH -J {c.name}
#SBATCH -p {c.script.queue}
#SBATCH --nodes=1
#SBATCH --ntasks-per-node={c.script.cores}
#SBATCH --mem={c.script.memory}G
#SBATCH --time={c.script.walltime}
#SBATCH -o {c.log_file}
{c.script.extra_args_str}

{{MAIN}}
"""

PBS_TEMP = """{c.script.shebang}
#PBS -N {c.name}
#PBS -q {c.script.queue}
#PBS -l nodes=1:ppn={c.script.cores},walltime={c.script.walltime}
#PBS -l vmem={c.script.memory}gb
#PBS -j oe
#PBS -o {c.log_file}
{c.script.extra_args_str}

cd $PBS_O_WORKDIR
{{MAIN}}
"""

MGR_SYS = {
    "":      ("perl",   "",        DFLT_TEMP),
    "slurm": ("sbatch", "#SBATCH", SLURM_TEMP),
    "pbs":   ("qsub",   "#PBS",    PBS_TEMP),
}

QUERY_TIMEOUT = 5

def eval_var(sc):
    for v, fn in VAR.items():
        sc = sc.replace("{{"+v+"}}", fn())
    return sc

def norm_wtime(wtime):
    wtime = wtime.replace('-', ':')
    *d, h, m, s = wtime.split(':')
    if len(d) > 1: raise ValueError(f"{wtime!r} is not a valid walltime")
    t = int(h)*60 + int(m)
    if len(d) == 1:
        t += int(d[0])*24*60
    tw = t - 3 # 3min for worker to quit, be polite
    return f"{t//60:02}:{t%60:02}:{s}", f"{tw//60:02}:{tw%60:02}:{s}"

def eval_defer(setup_cleanup):
    setup = setup_cleanup.split('\n')
    cleanup = []
    i = 0
    while i < len(setup):
        if not setup[i].startswith("defer "):
            i += 1
            continue
        cleanup.append(setup.pop(i)[6:])
    cleanup.reverse()
    return '\n'.join(setup), '\n'.join(cleanup)

def submit(scmd, sc, *args, env=None):
    with tempfile.NamedTemporaryFile(mode='w', dir=".", buffering=1) as f:
        f.write(sc)
        subprocess.run([scmd, f.name, *args], env={**os.environ, **env} if env else None)


click.option = functools.partial(click.option, show_default=True)

def global_options(fn=None, conf_mode=''):
    if fn is None:
        return functools.partial(global_options, conf_mode=conf_mode)
    @click.option('-c', '--config', default="grain.toml", envvar='GRAIN_CONFIG', help="Grain config file. Can be set by envar `GRAIN_CONFIG`")
    @functools.wraps(fn)
    def _wrapped(config, *args, **kwargs):
        return fn(conf=load_conf(config, conf_mode), *args, **kwargs)
    return _wrapped

def gen_script(conf, is_worker=False):
    scriptc = conf.script
    rmem = int(scriptc.memory/15*14)
    if is_worker:
        res = dict()
        if scriptc.cores or rmem:
            res["Node"] = dict(N=scriptc.cores, M=rmem)
        if scriptc.walltime:
            scriptc.walltime, rwalltime = norm_wtime(scriptc.walltime)
            res["WTime"] = dict(T=rwalltime, countdown=True)
        scriptc.res_str = json.dumps(dict(
            **res,
            **conf.res,
        ))
    scriptc.setup, scriptc.cleanup = eval_defer(scriptc.setup_cleanup)
    if conf.custom_system:
        scmd, dirc, temp = conf.custom_system.submit_cmd, conf.custom_system.directory, conf.custom_system.template
    else:
        try:
            scmd, dirc, temp = MGR_SYS[conf.system.lower()]
        except KeyError:
            print(f"Unknown conf.system: {conf.system}")
            sys.exit(1)
    scriptc.extra_args_str = '\n'.join(f"{dirc} {l}" for l in scriptc.extra_args)
    return scmd, temp

@click.group(help="CLI tool to manage Grain workers")
@click.version_option(__version__)
def main():
    pass

@main.command(help="Submit worker jobs (extra opts are forwarded to the submission command)",
              context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
@global_options(conf_mode='worker')
@click.option('-n', type=int, default=1, help="Number of worker jobs/nodes to submit")
@click.option('--dry', is_flag=True, help="Dry run. Print out the job submission script without submitting. This ignores `-n`.")
@click.pass_context
def up(ctx, conf, n, dry):
    if conf.specialized_type and load_mod(conf.specialized_type).GRAIN_SWORKER_CONFIG.get("BACKENDLESS", False):
        print(f"Specialized worker {conf.specialized_type!r} is backendless, so you don't need to run a worker.", file=sys.stderr)
        return
    scmd, temp = gen_script(conf, is_worker=True)
    sc = temp.replace("{{MAIN}}", WORKER_MAIN).format(c=conf)
    if dry:
        print(eval_var(sc))
        return
    for i in range(n):
        if i > 0: sleep(1)
        submit(scmd, eval_var(sc), *ctx.args)

@main.command(help="Start the head process")
@global_options(conf_mode='head')
@click.option('--local', is_flag=True, help="Run locally instead of submitting a job")
@click.option('--dry', is_flag=True, help="Dry run. Print out the job submission script without submitting")
def start(conf, local, dry):
    scmd, temp = gen_script(conf)
    conf.log_file = conf.main_log_file # We want to record Grain's log
    sc = temp.replace("{{MAIN}}", HEAD_MAIN).format(c=conf)
    if dry:
        print(eval_var(sc))
        return
    if local: # FIXME: relay SIGINT
        submit(conf.script.shebang[2:], eval_var(sc), env=dict(PBS_O_WORKDIR=os.getcwd()))
    else:
        submit(scmd, eval_var(sc))

@main.command(help="Overview of workers' status")
@global_options(conf_mode='worker')
def ls(conf):
    async def _ls(head):
        with _handle_connection_error(head):
            async with trio.open_nursery() as _n, \
                       SocketChannel(f"{head}", dial=True, _n=_n) as c:
                await c.send(dict(cmd="STA"))
                print((await c.receive())['result'].decode())
    trio.run(_ls, conf.cli_dial)

@main.command(help="Quit workers, by name / glob pattern")
@global_options(conf_mode='worker')
@click.option('-f', '--force', is_flag=True, help="Kill the worker(s) even if it still have running jobs")
@click.argument('worker_name_pattern')
def quit(conf, force, worker_name_pattern):
    async def _unreg(head, name):
        with _handle_connection_error(head):
            await notify(f"{head}", dict(cmd="UNR" if force else "TRM", name=name), seg=True)
    trio.run(_unreg, conf.cli_dial, worker_name_pattern)

@main.command(help="Enable / change worker autoscaling")
@global_options(conf_mode='worker')
@click.argument('swarm_size', type=int)
def scale(conf, swarm_size):
    async def _scale(head, swarm_size):
        with _handle_connection_error(head):
            await notify(f"{head}", dict(cmd="SCL", name=str(swarm_size)), seg=True)
    trio.run(_scale, conf.cli_dial, swarm_size)

@contextmanager
def _handle_connection_error(url):
    try:
        with trio.fail_after(QUERY_TIMEOUT):
            yield
    except (OSError, trio.TooSlowError):
        print(f"unable to contact head through {url}")

if __name__=="__main__":
    main()



from grain.delayed import delayed, each, run
from grain.resource import Memory
import trio
from functools import partial

import pytest
from . import asyncify, N

@delayed
async def nop():
    await trio.sleep(0)

async def _serial0():
    """{N} empty jobs with ZERO res, serial"""
    for _ in range(N):
        await nop()

async def _parall0():
    """{N} empty jobs with ZERO res, parall"""
    await each([nop() for _ in range(N)])

async def _serialx():
    """{N} empty jobs with some res, serial"""
    for _ in range(N):
        await (nop@Memory(0))()

async def _parallx():
    """{N} empty jobs with some res, parall"""
    await each([(nop@Memory(0))() for _ in range(N)])

async def _main(main, benchmark):
    await (asyncify(benchmark))(main)
@pytest.mark.parametrize(
    "main",
    [_serial0, _parall0, _serialx, _parallx],
)
def test_delayed(benchmark, main):
    run(
        partial(_main, main, benchmark),
        config_file=False,
        local=Memory(0),
    )



import trio

from copy import deepcopy
from math import inf as INFIN
from functools import partial
import traceback
from contextlib import redirect_stdout, redirect_stderr, ExitStack
import sys
from fnmatch import fnmatchcase
import logging
logger = logging.getLogger(__name__)

from .contextvar import GVAR
from .util import timeblock, pickle_dumps, pickle_loads, WaitGroup, nullacontext, optional_cm, make_prependable, load_contextmod, load_mod, as_daemon
from .resource import Resource, ZERO, Reject
from .pair import SocketChannel, SocketChannelAcceptor, notify
from .subproc import subprocess_pool_daemon, BrokenSubprocessError
from .stat import log_event, log_timestart, log_timeend, stat_logger, ls_worker_res, reg_probe, collect_probe

FULL_HEALTH = 3
STATSPAN = 15 # minutes
HEARTBEAT_INTERVAL, HEARTBEAT_TOLERANCE = 10, 3 # 10s * 3

class GrainRemote:
    def __init__(self, addr, res):
        self._c = None
        self.res = res
        self.name = addr
        self.health = FULL_HEALTH
        self.wg = WaitGroup()
        self.resultq = {}
    async def _loop(self, task_status=trio.TASK_STATUS_IGNORED):
        async def heartbeat_s(c):
            while True:
                await c.try_send(dict(cmd="HBT"))
                await trio.sleep(HEARTBEAT_INTERVAL)
        async with self._c, \
                   trio.open_nursery() as _n:
            task_status.started()
            _n.start_soon(heartbeat_s, self._c)
            while True:
                x = None
                try:
                    with trio.move_on_after(HEARTBEAT_INTERVAL*HEARTBEAT_TOLERANCE):
                        x = await self._c.receive()
                except trio.EndOfChannel:
                    break
                if x is None:
                    logger.warning(f"remote {self.name} heartbeat response timeout")
                    break
                elif x == {'cmd':'HBT'}:
                    continue
                tid, ok, r = x['tid'], not x['exception'], x['result']
                rq = self.resultq.get(tid)
                if rq:
                    await rq.send((ok, pickle_loads(r)))
                else:
                    log_event("late_response")
                    logger.warning(f"remote {self.name} received phantom job {tid}'s result")
            _n.cancel_scope.cancel()
        # well, we just let the remote decide when to leave
        #assert self._c.is_clean is False # NOTE: P2P conn doesn't get EOF when the other end quit, we need to close on our end
        # In case of connection lost, dismiss all pending jobs
        for rq in list(self.resultq.values()): # frozen
            await rq.aclose()
    async def connect(self, _n):
        self._c = SocketChannel(f"tcp://{self.name}:4242", dial=True, _n=_n)
        await _n.start(self._loop)
    async def execf(self, tid, res, fn):
        with self.wg:
            self.resultq[tid], rq = trio.open_memory_channel(0)
            try:
                await self._c.send(dict(tid=tid, res=res, func=pickle_dumps(fn)))
                with optional_cm(trio.fail_after, getattr(res,'T',-180)+180): # 3min grace period
                    ok, r = await rq.receive()
            except (trio.ClosedResourceError, trio.EndOfChannel):
                # TODO: dedicated error class?
                ok, r = False, ("",RuntimeError(f"remote {self.name} closed connection unexpectedly"))
            except trio.TooSlowError:
                log_event("lost_or_late_response")
                ok, r = False, ("",trio.TooSlowError(f"remote {self.name} lost track of job {tid}"))
            if ok: self.health = FULL_HEALTH
            del self.resultq[tid]
            return ok, r
    async def aclose(self):
        await self._c.try_send(dict(cmd="FIN")) # fails if KI or connection lost
        logger.info(f"worker {self.name} starts cleaning up")
        await self.wg.wait()
        await self._c.aclose() # for P2P connection
        logger.info(f"worker {self.name} clear")

class GrainReverseRemote(GrainRemote):
    """A Grain remote for which the worker dial in
    (e.g. in the case when worker doesn't have accessible
    port, but is able to access the head.)
    The registration message (i.e. (vaddr, res)) should
    be sent as the first message.
    See `GrainManager.worker_manager`.
    """
    def __init__(self, _c, vaddr, res):
        GrainRemote.__init__(self, vaddr, res)
        self._c = _c
    async def connect(self, _n):
        _n.start_soon(self._loop)

GVAR.instance = "N/A"
class _cobj:
    def __init__(self, fobj, dobj):
        self._h, self._dh = fobj, dobj
    def __getattr__(self, attr):
        if GVAR.instance.endswith("(local)"):
            return getattr(self._h, attr)
        return getattr(self._dh, attr)
class GrainPseudoRemote:
    def __init__(self, name, res, out="", cmod=nullacontext()):
        self.res = res
        self.name = name
        self.health = FULL_HEALTH
        self.wg = WaitGroup()
        self.cg = set()
        self.redi_cm = ExitStack()
        if out:
            outh = self.redi_cm.enter_context(open(out,'w'))
            self.redi_cm.enter_context(redirect_stdout(_cobj(outh,sys.stdout)))
            self.redi_cm.enter_context(redirect_stderr(_cobj(outh,sys.stderr)))
        self.cmod = cmod
        self.cmod_context = None
    async def _scope(self, task_status=trio.TASK_STATUS_IGNORED):
        with trio.CancelScope() as cs, \
             self.redi_cm:
            self.cg.add(cs)
            async with self.cmod as self.cmod_context:
                task_status.started()
                await trio.sleep_forever()
    async def connect(self, _n):
        await _n.start(self._scope)
    async def execf(self, tid, res, fn):
        with self.wg, \
             trio.CancelScope() as cs:
            self.cg.add(cs)
            GVAR.res = res
            GVAR.instance = self.name
            GVAR.context = self.cmod_context
            try:
                with optional_cm(trio.fail_after, getattr(res,'T',0)):
                    ok, r = True, await fn()
                    self.health = FULL_HEALTH
            except BaseException as e:
                ok, r = False, (traceback.format_exc(), e)
            self.cg.remove(cs)
            return ok, r
    async def aclose(self):
        for cs in self.cg:
            cs.cancel()
        logger.info(f"worker {self.name} starts cleaning up")
        await self.wg.wait()
        logger.info(f"worker {self.name} clear")

class GrainSpecializedRemote(GrainPseudoRemote):
    # TODO: heartbeat
    def __init__(self, name, res, _c, sw_kwargs):
        stype = sw_kwargs.pop("_stype")
        sw_mod = load_mod(stype)
        super().__init__(name, res, "", sw_mod.grain_context(**sw_kwargs))
        self._c = _c
    async def aclose(self):
        for cs in self.cg:
            cs.cancel()
        if self._c:
            await self._c.try_send(dict(cmd="FIN")) # fails if KI or connection lost
            logger.info(f"worker {self.name} starts cleaning up")
        else:
            logger.info(f"client for specialized worker {self.name} starts cleaning up")
        await self.wg.wait()
        if self._c:
            await self._c.aclose() # for P2P connection
            logger.info(f"worker {self.name} clear")
        else:
            logger.info(f"client for specialized worker {self.name} clear")


class GrainManager:
    """Manage workers and resources.
    """
    def __init__(self, pool_init, _n, temperr, persistent, listen_addr, sworker_config):
        self.pool = pool_init
        self._n = _n
        self.temperr = {*set(temperr), BrokenSubprocessError}
        self.persistent = persistent
        self.listen_addr = listen_addr
        self.sworker_config = sworker_config
        self.cond_res = trio.Condition()
        self.soacceptor = None

    async def schedule(self, res): # naive scheduling alg: greedy
        async with self.cond_res:
            while True:
                for w in self.pool:
                    if w.res >= res: return w.res.alloc(res), w
                await self.cond_res.wait()

    async def health_check(self, w, tb, err):
        async with self.cond_res:
            w.health -= 1 if type(err) in self.temperr else INFIN
            if w.health <= 0 and w in self.pool:
                if type(err) not in self.temperr:
                    logger.error(tb)
                if w.name.endswith('(local)'): raise RuntimeError("local worker quits")
                logger.warning(f"quit worker {w.name} due to poor health {w.health}")
                await self.unregister(w.name, locked=True)

    async def dealloc(self, w, res):
        if w.health > 0:
            async with self.cond_res:
                w.res.dealloc(res)
                self.cond_res.notify()

    async def register(self, w, _n):
        await w.connect(_n)
        async with self.cond_res:
            self.pool.append(w)
            self.cond_res.notify()
    def _match_workers(self, p):
        # results are only meaningful within cond_res
        return [(i,w) for i,w in enumerate(self.pool)
                if fnmatchcase(w.name,p) and not w.name.endswith('(local)')]
    async def unregister(self, pattern, locked=False):
        async with nullacontext() if locked else self.cond_res:
            for i,w in reversed(self._match_workers(pattern)):
                logger.info(f"worker {w.name} is quitting now")
                await w.aclose() # notify exit and reschedule pending jobs
                if not self.persistent:
                    raise RuntimeError(f"worker {w.name} exits, abort.")
                del self.pool[i]
    async def terminate(self, pattern):
        async def _wait_n_unreg(w):
            await w.wg.wait() # wait till no job is running
            await self.unregister(w.name) # `self.pool` might have changed, do another lookup
        async with self.cond_res:
            for _,w in self._match_workers(pattern):
                if type(w.res) is Reject: continue
                logger.info(f"worker {w.name} is going to leave")
                w.res = Reject(w.res)
                self._n.start_soon(_wait_n_unreg, w)

    async def worker_manager(self, registered_stype, task_status=trio.TASK_STATUS_IGNORED):
        async with trio.open_nursery() as _n, \
                   SocketChannelAcceptor(self.listen_addr, _n=_n) as self.soacceptor:
            for w in self.pool:
                await w.connect(_n)
            task_status.started()
            async for _c in self.soacceptor:
                try:
                    msg = await _c.receive()
                except trio.EndOfChannel:
                    continue
                cmd = msg['cmd']
                if cmd == "CON": # connect
                    vaddr, res = msg['name'], msg['res']
                    logger.info(f"worker {vaddr} joined with {res}")
                    await self.register(GrainReverseRemote(_c, vaddr, res), _n)
                    continue
                elif cmd == "SRG": # specialized register
                    name, res, kwargs = msg['name'], msg['res'], msg['obj']
                    logger.info(f"worker {name} joined with {res}")
                    if kwargs['_stype'] not in registered_stype:
                        raise ValueError(f"Unknown specialized worker type {kwargs['_stype']}")
                    await self.register(GrainSpecializedRemote(name, res, _c, kwargs), _n)
                    continue
                async with _c: # The following are ephemeral cmds
                    if cmd == "REG": # register
                        addr, res = msg['name'], msg['res']
                        logger.info(f"worker {addr} joined with {res}")
                        await self.register(GrainRemote(addr, res), _n)
                    elif cmd == "UNR": # unregister
                        pattern = msg['name']
                        await self.unregister(pattern)
                    elif cmd == "STA": # statistics
                        sta = ls_worker_res(self.pool) + '\n' + \
                              collect_probe()
                        await _c.try_send(dict(result=sta.encode()))
                    elif cmd == "TRM": # graceful termination
                        pattern = msg['name']
                        await self.terminate(pattern)
                    else:
                        logger.warning(f"worker manager received unknown command {cmd} from {_c.host}")

    async def __aenter__(self):
        registered_stype, _, self.backendless_sworker_n = await self._n.start(
            backendless_sworker_manager,
            self.sworker_config, self.listen_addr,
        )
        await self._n.start(self.worker_manager, registered_stype)

    async def aclose(self):
        with trio.move_on_after(10) as cleanup_scope: # 10s cleanup
            cleanup_scope.shield = True
            self.backendless_sworker_n.cancel_scope.cancel()
            for w in self.pool:
                await w.aclose()
            await self.soacceptor.aclose()

    async def __aexit__(self, *exc):
        await self.aclose()
        return False


async def backendless_sworker_manager(sworker_config, manager_api, task_status=trio.TASK_STATUS_IGNORED):
    registered_stype = set()
    backendless_stype = set()
    async with trio.open_nursery() as backendless_sworker_n:
        for sw_conf in sworker_config:
            stype = sw_conf.specialized_type
            registered_stype.add(stype)
            sw_mod = load_mod(stype)
            if not sw_mod.GRAIN_SWORKER_CONFIG.get("BACKENDLESS", False):
                continue
            backendless_stype.add(stype)
            # Backendless sworker's `grain_run_sworker` should be a trivial context for returning `sw_info`
            sw_info = await backendless_sworker_n.start(as_daemon, sw_mod.grain_run_sworker())
            resd = dict()
            if sw_conf.script.cores or sw_conf.script.memory:
                resd["Node"] = dict(N=sw_conf.script.cores, M=sw_conf.script.memory)
            if sw_conf.script.walltime:
                resd["WTime"] = dict(T=sw_conf.script.walltime, countdown=True)
            res = Resource.from_dict({
                **resd,
                **sw_conf.res,
            })
            await notify(
                manager_api,
                dict(cmd="SRG", name=sw_info.pop("name"), res=res, obj=dict(_stype=stype, **sw_info)),
                seg=True,
            )
        task_status.started((registered_stype, backendless_stype, backendless_sworker_n))


class GrainExecutor:
    """There are two ways to use GrainExecutor: sync and async:
    sync: TODO
    async: TODO
    """
    def __init__(
        self,
        waddrs=(),        # list of initial passive workers' addresses
        rpw=ZERO,         # resource per worker for the initial passive workers
        _n=None,          # (in async mode) external trio.Nursery for self's and manager's run loop
        nolocal=False,    # run jobs on local or not
        temporary_err=(), # exceptions that's not critical to shutdown a worker
        reschedule=True,  # if False, abort on any exception
        persistent=True,  # if False, abort on any worker's exit
        config=None,      # grain's head config
        stat_tag=lambda res: str(getattr(res,'T',"")), # define how timestat is categorized by resource
        sworker_config=(), # grain's specialized worker config
     ):
        self.push_job, self.pull_job = trio.open_memory_channel(INFIN)
        self.push_job = make_prependable(self.push_job)
        self.push_result, self.resultq = trio.open_memory_channel(INFIN)
        self.jobn = 0
        self.results = []
        self._n = _n
        self._wg = WaitGroup() # track the entire lifetime of each job
        self.reschedule = reschedule
        self.mgr = GrainManager(
            [GrainPseudoRemote(
                f"{trio.socket.gethostname()}(local)",
                deepcopy(rpw if not nolocal else ZERO),
                config.log_file,
                load_contextmod(config.contextmod)()
            )] + [GrainRemote(a, deepcopy(rpw)) for a in waddrs],
            self._n, temporary_err, persistent, config.listen, sworker_config)
        reg_probe("queued jobs", lambda: len(self.push_job._state.data))
        self.stat_tag = stat_tag

    async def asubmit(self, res, fn, *args, **kwargs):
        self._wg.add()
        tid = self.jobn = self.jobn+1
        await self.push_job.send((tid, res, partial(fn, *args, **kwargs)))
        return tid
    def submit(self, res, fn, *args, **kwargs):
        self._wg.add()
        tid = self.jobn = self.jobn+1
        self.push_job.send_nowait((tid, res, partial(fn, *args, **kwargs)))
        return tid
    def _submit(self, tid, res, fn): # for internal use
        self._wg.add()
        self.push_job.send_nowait((tid, res, fn))
    def resubmit(self, tid, res, fn):
        self._wg.add()
        self.push_job.cutin_nowait((tid, res, fn)) # prepend the queue

    async def __task_with_res(self, tid, res, w, fn):
        try:
            tag = self.stat_tag(res)
            if tag: log_timestart(tag, tid)
            ok, r = await w.execf(tid, res, fn)
            if tag: log_timeend(tag, tid)
            if ok:
                log_event("completed")
                self.push_result.send_nowait((tid, r))
            else:
                log_event("error")
                tb, err = r
                if not self.reschedule:
                    raise RuntimeError(f"worker {w.name}'s task {fn} raises {err.__class__.__name__}: {err}, abort.")
                logger.error(f"worker {w.name}'s task {fn} raises {err.__class__.__name__}: {err}, going to reschedule it...")
                await self.mgr.health_check(w, tb, err)
                self.resubmit(tid, res, fn) # preserve tid
            await self.mgr.dealloc(w, res)
        finally:
            self._wg.done()

    async def sealer(self):   # out of the executor scope => end of submission
        await self._wg.wait() # no pending task => end of resubmission
        await self.push_job.aclose()

    async def run(self, task_status=trio.TASK_STATUS_IGNORED):
        res = None
        reg_probe("next pending job's res", lambda: res)
        with timeblock("all jobs"):
            async with self.mgr, \
                       stat_logger(STATSPAN), \
                       self.push_result, \
                       trio.open_nursery() as _n, \
                       self.pull_job:
                task_status.started()
                async for tid, res, fn in self.pull_job:
                    res, w = await self.mgr.schedule(res)
                    _n.start_soon(self.__task_with_res, tid, res, w, fn)

    async def __aenter__(self):
        await self._n.start(self.run)
        return self
    async def __aexit__(self, *exc):
        if any(exc):
            return False
        self._n.start_soon(self.sealer)

    async def run_till_finish(self):
        async with trio.open_nursery() as self._n:
            self.mgr._n = self._n
            await self._n.start(self.run)
            self._n.start_soon(self.sealer)
        self.results = [None] * self.jobn
        async with self.resultq:
            async for i, r in self.resultq:
                self.results[i-1] = r
    def __enter__(self):
        return self
    def __exit__(self, *exc):
        if any(exc):
            return False
        trio.run(self.run_till_finish)



import time
from collections import defaultdict
from math import inf as INFIN
from statistics import fmean as mean, stdev

__all__ = ["Resource", "ZERO", "Cores", "Memory", "Node", "WTime", "Token", "Capacity",
           "ONE_INSTANCE", "Reject", "REJECT"]

class Resource:
    def _request(self, res):
        """Return true if it is possible to alloc(res)
        """
    def _alloc(self, res):
        """Return allocated resources. Allocated resources
        should also be able to be used as request resources (i.e.
        ``B.request(A.alloc(some_res))``), but vice versa does not
        necessary need to be true.
        """
    def _dealloc(self, res):
        pass
    def _repr(self):
        """``__repr__``, for concatenating with other resources
        """
    def _stat(self):
        """Return (p,q) where p and q are `int`, and p/q
        represents the percentage resource availability.
        """

    def __init__(self, init=True):
        self.__resm = { self.__class__.__name__: self } if init else {}
    def __repr__(self):
        if not self.__resm: return "Ø"
        return ' & '.join(map(lambda x:x._repr(), self.__resm.values()))
    def __and__(self, other):
        if any(k in self.__resm for k in other.__resm.keys()):
            raise ValueError("Combining resources of the same type is not supported")
        r = Resource(init=False)
        r.__resm = { **self.__resm, **other.__resm }
        return r
    def __getattr__(self, attr):
        if attr == "__setstate__": # pickle hook
            raise AttributeError
        for v in self.__resm.values():
            if attr in v.__dict__: # static lookup only
                return v.__dict__[attr]
        raise AttributeError(f"Resource object {self} has no attribute {attr}")
    def __ge__(self, other): # other is affordable
        return self.request(other)
    def __gt__(self, other):
        return not other.request(self)
    def __le__(self, other):
        return other.request(self)
    def __lt__(self, other):
        return not self.request(other)
    def __eq__(self, other):
        return self.request(other) and other.request(self)

    def request(self, res):
        for k,v in res.__resm.items():
            if (k not in self.__resm) or \
               (not self.__resm[k]._request(v)):
                return False
        return True
    def alloc(self, res):
        if any(k not in self.__resm for k in res.__resm.keys()):
            raise ValueError(f"Cannot alloc: resource {self} does not have the same shape as {res}")
        r = Resource(init=False)
        try:
            for k,v in res.__resm.items():
                r &= self.__resm[k]._alloc(v)
        except: # rewind side-effects
            self.dealloc(r)
            raise
        return r
    def dealloc(self, res):
        if any(k not in self.__resm for k in res.__resm.keys()):
            raise ValueError(f"Cannot dealloc: resource {self} does not have the same shape as {res}")
        for k,v in res.__resm.items():
            self.__resm[k]._dealloc(v)
    def stat(self):
        return {k:v._stat() for k,v in self.__resm.items()}
    def encode_msgp(self):
        return {k:v._encode_msgp() for k,v in self.__resm.items()}

    @classmethod
    def from_dict(cls, resd):
        from . import resource
        res = cls(init=False)
        for rn, rargs in resd.items():
            res &= getattr(resource, rn)(**rargs)
        return res



ZERO = Resource(init=False)


class Cores(Resource): # CPU cores

    def __init__(self, N):
        super().__init__()
        if type(N) is int:
            self.c = set(range(N))
            self.N = N
        else: # assume iterable
            self.c = set(N)
            self.N = len(N)
    def _repr(self):
        if self.N == 0: return "CPU_Cores([])"
        first, *c = sorted(self.c)
        con = [[first,first]]
        for x in c:
            if x == con[-1][1]+1: con[-1][1] = x
            else: con.append([x,x])
        clist = ','.join((str(x) if x==y else f"{x}-{y}") for x,y in con)
        return f"CPU_Cores([{clist}])"

    # For request/alloc we only care about the count but not identities
    def _request(self, res):
        return len(self.c) >= res.N

    def _alloc(self, res):
        if not self._request(res):
            raise ValueError(f"{self} cannot allocate {res.N} core(s)")
        a = list(self.c)[:res.N]
        self.c -= set(a)
        return Cores(a)

    def _dealloc(self, res):
        self.c |= res.c

    def _stat(self):
        return len(self.c), self.N

    def _encode_msgp(self):
        return dict(N=sorted(self.c))


class Memory(Resource): # vmem

    def __init__(self, M):
        super().__init__()
        self.m = M
        self.M = M
    def _repr(self):
        return f"Memory({self.m}GB)"

    def _request(self, res):
        return self.m >= res.m

    def _alloc(self, res):
        if not self._request(res):
            raise ValueError(f"{self} cannot allocate {res}")
        res = Memory(res.m)
        self.m -= res.m
        return res

    def _dealloc(self, res):
        self.m += res.m

    def _stat(self):
        return self.m, self.M

    def _encode_msgp(self):
        return dict(M=self.M)


def Node(N, M):
    return Cores(N) & Memory(M)


TIMESTAT_NSAMPLE = 8
TIMESTAT = defaultdict(list)
TIMEINFR = {}
class WTime(Resource): # walltime, not restorable

    def __init__(self, *, T=None, softT=None, group=None, countdown=False): # countdown ? allocer : requester
        super().__init__()
        if isinstance(T, str):
            *d, h, m, s = map(int, T.replace('-',':').split(':'))
            T = 86400*(d or [0])[-1] + 3600*h + 60*m + s
        elif isinstance(T, int) and T > 10*365*86400:
            T = INFIN
        self.deadline = time.time() + T if countdown else 0
        self.T = T
        self.softT = softT or T
        self.group = group
    def _repr(self):
        if self.group:
            if self.group not in TIMEINFR:
                return f"Walltime(group={self.group})"
            t = TIMEINFR[self.group].T
        else:
            t = max(int(self.deadline - time.time()), 0) if self.deadline else self.T
        return f"Walltime({t//3600:02}:{t%3600//60:02}:{t%60:02})"

    def _request(self, res):
        if res.group:
            if res.group not in TIMEINFR:
                return True
            res = TIMEINFR[res.group]
        return self.deadline - time.time() >= res.softT # loose condition

    def _alloc(self, res):
        if not self._request(res):
            raise ValueError(f"{self} cannot allocate {res}")
        if res.group:
            if res.group in TIMEINFR:
                return TIMEINFR[res.group]
            res = WTime(T=0, group=res.group, countdown=True) # store current time
            res.T = INFIN # ... while set no time limit
        return res

    def _dealloc(self, res):
        if res.group and res.group not in TIMEINFR:
            if (delt := time.time() - res.deadline) < 10:
                return
            (s := TIMESTAT[res.group]).append(delt)
            if len(s) == TIMESTAT_NSAMPLE:
                tm = mean(s)
                ts = stdev(s, tm)
                TIMEINFR[res.group] = WTime(T=round(tm+5*ts), softT=round(tm+2*ts))
                del TIMESTAT[res.group]

    def _stat(self):
        return 1, 1

    def _encode_msgp(self):
        if self.group:
            return dict(group=self.group)
        if self.deadline:
            return dict(T=int(self.deadline-time.time()), countdown=True)
        else:
            return dict(T=int(self.T), softT=int(self.softT))


class Token(Resource):

    def __init__(self, token):
        super().__init__()
        self.token = token
    def _repr(self):
        return f"Token({self.token})"

    def _request(self, res):
        return self.token == res.token

    def _alloc(self, res):
        if not self._request(res):
            raise ValueError(f"{self} does not match {res}")
        return res

    def _dealloc(self, res):
        pass

    def _stat(self):
        return 1, 1


class Capacity(Resource):

    def __init__(self, V):
        super().__init__()
        self.V = self.v = V
    def _repr(self):
        return f"Capacity({self.v})" if self.v != -1 else "ONE"

    def _request(self, res):
        return self.v > 0

    def _alloc(self, res):
        if not self._request(res):
            raise ValueError(f"{self} has no capacity")
        self.v -= 1
        return res

    def _dealloc(self, res):
        self.v += 1

    def _stat(self):
        return self.v, self.V

    def _encode_msgp(self):
        return dict(V=self.V)

ONE_INSTANCE = Capacity(-1) # for requester
# This is not ideal as ONE_INSTANCE != ONE_INSTANCE


class Reject(Resource):
    """Similar to `ZERO`, but uncondionally suppress
    all request/alloc and absorb dealloc
    """
    def __init__(self, wrap=ZERO):
        super().__init__(init=False)
        self.__self = wrap
    def __repr__(self):
        return "N/A"
    def __and__(self, other):
        return self

    def request(self, res):
        return False
    def alloc(self, res):
        raise ValueError("No resource is or will be available")
    def dealloc(self, res):
        self.__self.dealloc(res) # Blackhole
    def stat(self):
        return self.__self.stat()

    def eject(self):
        return self.__self

REJECT = Reject()




"""
A dumb backendless specialized worker implementation for testing purpose.
By setting `worker.specialized_type = "grain.local_sworker"`,
tasklets are executed locally, subjected to resource constraint like `worker.res`.
Everything works similarly as if running `grain.delayed.run(local=...)`,
except that the Gnaw executor is being used.
"""
from contextlib import asynccontextmanager


GRAIN_SWORKER_CONFIG = dict(
    BACKENDLESS=True,
)


@asynccontextmanager
async def grain_context():
    yield None


@asynccontextmanager
async def grain_run_sworker():
    yield dict(name="local")




import trio

from math import inf as INFIN
from functools import partial
from urllib.parse import urlparse, parse_qsl

from .conn import open_tcp_stream_to_head, serve_tcp_p2p, make_ipaddr
from .conn_msgp import iter_packet, send_packet
from .conn2 import open_tcp_edge_stream, serve_tcp_edge

async def notify(url, msg, seg=False): # TODO: retry until connected
    """Open a connection, send `msg`, then close the
    connection immediately.
    To be used with `listen_signal`.
    Or set `seg=True` to talk to `SocketChannel`.
    """
    proto, host, port, opts = parse_url(url)
    async with await open_tcp_stream(proto, host, port, opts) as s:
        if seg:
            await send_packet(s, msg)
        else:
            await s.send_all(msg)

class SocketReceiveChannel(trio.abc.ReceiveChannel):
    def __init__(self, url, _n):
        self.proto, self.host, self.port, self.opts = parse_url(url)
        self.in_s, self.in_r = trio.open_memory_channel(INFIN)
        self._n = _n # for any background loops
        self._cs = trio.CancelScope() # for terminating all loops while keep _n intact
    async def receive(self):
        return await self.in_r.receive()
    async def aclose(self):
        async with self.in_r, self.in_s:
            self._cs.cancel()

class EphemeralSocketReceiveChannel(SocketReceiveChannel):
    """Accept any connection. For each connection
    receive a single fragment (trailed by EOF) of data,
    then immediately close the connection.
    `receive` returns item type of
    `((addr: str, port: int), data: byte)`.
    To be used with `notify`.
    """
    async def __aenter__(self):
        await self._n.start(partial(serve_tcp, self.proto, self._handler,
                                    self.port, host=self.host, opts=self.opts, cs=self._cs))
        return self
    async def _handler(self, s):
        async with s: # one msg per connection
            msg = b''
            async for x in s: msg += x
            self.in_s.send_nowait((s.socket.getpeername(), msg))

listen_signal = EphemeralSocketReceiveChannel


class SocketChannel(trio.abc.SendChannel, SocketReceiveChannel):
    """Dial to one endpoint / accept one connection in
    order to setup a single duplex connection.
    Alternatively, pass in a connected socket stream.
    `receive` returns item type of `data: byte`
    """
    def __init__(self, url, _n=None, listen=False, dial=False, _so=None):
        SocketReceiveChannel.__init__(self, url, _n)
        self._send_lock = trio.Lock()
        assert (listen != dial and _n is not None) or _so is not None
        self._so, self.listen, self.dial = _so, listen, dial
        self.is_clean = True
    async def __aenter__(self):
        if self.listen:
            await self._n.start(partial(serve_tcp, self.proto, self._handler,
                                        self.port, host=self.host, opts=self.opts, cs=self._cs))
        elif self.dial:
            self._so = await open_tcp_stream(self.proto, self.host, self.port, self.opts)
            self._n.start_soon(self._handler_standalone, self._so)
        return self
    async def _handler(self, s):
        self._so = s
        async with s:
            async for p in iter_packet(s):
                self.in_s.send_nowait(p)
            else: # socket terminated by remote
                self.is_clean = False
                await self.in_s.aclose()
    async def _handler_standalone(self, s):
        # for listeners, we cancel the accept loop, while here we cancel the only stream
        with self._cs:
            await self._handler(s)
    async def send(self, data):
        if not self._so: raise TypeError("socket not connected")
        if not data: return
        async with self._send_lock:
            await send_packet(self._so, data)
    async def try_send(self, data):
        try:
            await self.send(data)
        except trio.ClosedResourceError:
            pass


class SocketChannelAcceptor(SocketReceiveChannel):
    """Listener-like object that wraps the accepted TCP socket
    connections with `SocketChannel`.
    It also implements `trio.abc.ReceiveChannel`, so it can be
    used as a generator of `SocketChannel`. Note that the
    `SocketChannel` it returns is already connected and has its
    internal loop run under an external nursery. This implies
    1) its `__aenter__` is a NOP, b/c it is ready to be used,
    and 2) calling `SocketChannelAcceptor.aclose` will not
    close any child `SocketChannel`.
    """
    async def __aenter__(self):
        await self._n.start(partial(serve_tcp, self.proto, self._handler, self.port, host=self.host,
                                    handler_nursery=self._n, opts=self.opts, cs=self._cs))
        return self
    async def _handler(self, so):
        # TODO: support unix server
        host, port, *_ = so.socket.getpeername()
        c = SocketChannel(f"tcp://{make_ipaddr(host, port)}", _so=so)
        self.in_s.send_nowait(c)
        await c._handler_standalone(so) # to have the same lifetime as c
    accept = SocketReceiveChannel.receive


async def open_tcp_stream(proto, host, port, opts=None):
    if proto == "tcp":
        return await trio.open_tcp_stream(host, port, happy_eyeballs_delay=INFIN)
    elif proto == "edge":
        return await open_tcp_edge_stream(edge_file=host)
    elif proto == "bridge":
        assert opts
        return await open_tcp_stream_to_head(bridge=(host, port), **opts)
    elif proto == "unix":
        return await trio.open_unix_socket(host)

# adapted from `trio.serve_tcp`, adding a CancelScope that terminates the listeners
async def serve_tcp(proto, handler, port, cs, *, host=None, handler_nursery=None, opts=None, task_status=trio.TASK_STATUS_IGNORED):
    with cs:
        if proto == "tcp":
            await trio.serve_tcp(handler, port, host=host, handler_nursery=handler_nursery, task_status=task_status)
        elif proto == "edge":
            await serve_tcp_edge(handler, edge_file=host, handler_nursery=handler_nursery, task_status=task_status)
        elif proto == "bridge":
            assert opts
            await serve_tcp_p2p(handler, bridge=(host,port), handler_nursery=handler_nursery, task_status=task_status, **opts)
        elif proto == "unix":
            raise NotImplementedError("Unix domain socket server is not supported")

def parse_url(url):
    u = urlparse(url)
    if u.scheme == "tcp":
        return u.scheme, u.hostname, u.port, {}
    elif u.scheme == "edge":
        return u.scheme, u.path, None, {}
    elif u.scheme == "bridge":
        assert u.username, "a session key must be specified for the bridge protocol"
        return u.scheme, u.hostname, u.port, { "key": u.username, **dict(parse_qsl(u.query)) }
    elif u.scheme == "unix":
        return u.scheme, u.path, None, {}
    else:
        raise ValueError(f"Unsupported protocol {u.scheme!r}")



import trio

import traceback
import argparse
import time
import json
from math import inf as INFIN
try:
    from exceptiongroup import BaseExceptionGroup  # for Python < 3.11
except ImportError:
    pass
import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

from .contextvar import GVAR
from .resource import Resource
from .pair import notify, SocketChannel
from .util import pickle_dumps, pickle_loads, load_contextmod, load_mod, nullacontext, as_daemon

cs_pool = {}

async def exerf(tid, res, func, so):
    GVAR.res = res
    try:
        with trio.fail_after(getattr(res, 'T', INFIN)) as cs:
            cs_pool[tid] = cs
            exception, r = "", await func()
            del cs_pool[tid]
    except BaseException as e:
        if cs_pool.pop(tid, None) is None:
            e = UserCancelled() # task indicated as no longer needed; to be discarded
        else:
            e = map_exc(e, (lambda e_: WorkerCancelled() if type(e_) is trio.Cancelled else e_))
        exception, r = e.__class__.__name__, (traceback.format_exc(), e)
    finally:
        with trio.move_on_after(3) as cleanup_scope:
            cleanup_scope.shield = True
            await so.send(dict(tid=tid, exception=exception, result=pickle_dumps(r)))

NO_NEXT_URL = "NO_NEXT_URL"

async def grain_worker(RES, url, maybe_sworker_mod=None):
    timesc = nullacontext()
    if hasattr(RES, 'T'):
        timesc = trio.fail_after(max(int(RES.deadline - time.time()), 0))
    passive = not (url and RES)
    if passive:
        sockopt = dict(url="tcp://:4242", listen=True)
        logger.info("passive mode enabled")
    else:
        sockopt = dict(url=url, dial=True) # TODO: connection timeout
        logger.info(f"connecting to head using {url!r}...")
    async with trio.open_nursery() as _cn, \
               SocketChannel(**sockopt, _n=_cn) as so, \
               trio.open_nursery() as _n:
        if not passive:
            if maybe_sworker_mod:
                sworker_info = await _n.start(as_daemon, maybe_sworker_mod.grain_run_sworker())
                name = GVAR.instance = sworker_info.pop("name")
                await so.send(dict(cmd="SRG", name=name, res=RES, obj=dict(
                   _stype=carg.sworker,
                   **sworker_info,
                )))
            else:
                await so.send(dict(cmd="CON", name=GVAR.instance, res=RES))
        logger.info("worker launched")
        try:
            with timesc:
                async for msg in so:
                    if 'cmd' not in msg:
                        tid, res, fn = msg['tid'], msg['res'], pickle_loads(msg['func'])
                        _n.start_soon(exerf, tid, res, fn, so)
                        continue
                    cmd = msg['cmd']
                    if cmd == "FIN": # end of queue / low health / server error
                        logger.info("received FIN from head, worker exits")
                        return NO_NEXT_URL
                    elif cmd == "HBT": # heartbeat
                        await so.send(dict(cmd="HBT"))
                        continue
                    elif cmd == "CAN": # cancel
                        for tid in map(int, msg['name'].split(',')):
                            if (cs := cs_pool.pop(tid, None)) is not None:
                                cs.cancel()
                    elif cmd == "REC": # reconnect
                        return msg['name']
                else:
                    logger.warning("connection to head lost, worker exits")
                    so.send = anop
                    return url # connection loss, reconnect to the same addr?
        except BaseException as e:
            if passive:
                logger.info(f"interrupted by {e.__class__.__name__}")
                return NO_NEXT_URL # `GVAR.instance` might differ from head's record, so don't notify
            logger.info(f"interrupted by {e.__class__.__name__}, notify head")
            await notify(url, dict(cmd="UNR", name=GVAR.instance), seg=True)
            if (await so.receive()) != {'cmd': 'FIN'}: assert False
            logger.info("received FIN from head, worker exits")
            return NO_NEXT_URL
        finally:
            _n.cancel_scope.cancel()

async def __loop():
    RES = parse_res(carg.res)
    url = carg.url
    maybe_sworker_mod = load_mod(carg.sworker)
    while url != NO_NEXT_URL:
        async with load_contextmod(carg.context)():
            url = await grain_worker(RES, url, maybe_sworker_mod)

def parse_res(res_str):
    res_dict = json.loads(res_str.replace('\\n', '\n'))
    return Resource.from_dict(res_dict)

async def anop(*_): pass

class WorkerCancelled(BaseException):
    def __str__(self):
        return "Cancelled"

class UserCancelled(BaseException):
    def __str__(self):
        return "Cancelled"

def map_exc(e, fn):
    if isinstance(e, BaseExceptionGroup):
        return BaseExceptionGroup(e.message, [map_exc(e_, fn) for e_ in e.exceptions])
    return fn(e)

if __name__ == "__main__":
    argp = argparse.ArgumentParser(description="Worker instance for Grain")
    argp.add_argument('--url', default="", help="URL to connect to Grain's head instance")
    argp.add_argument('--res', '-r', default="None", help='the resource owned by the worker, in JSON '
                                                          '(e.g. \'{"Node": { "N": 16, "M": 32 }, "WTime": { "T": "1:00:00", "countdown": true }}\')')
    argp.add_argument('--context', default="", help="context module file")
    argp.add_argument('--sworker', default="", help="specialized worker module file")
    carg = argp.parse_args()

    GVAR.instance = trio.socket.gethostname()
    trio.run(__loop)



from grain.remote_exer import RemoteExecutor
from grain.head import GrainExecutor
from grain.delayed import delayed, each, run
from grain.config import load_conf
from grain.resource import Cores
import trio
import sys
from io import StringIO

import pytest
from . import asyncify, N

async def nop():
    await trio.sleep(0)

async def _parallx(exer):
    for _ in range(N):
        exer.submit(Cores(0), nop)
    for _ in range(N):
        await exer.resultq.receive()

CONFIG_STR = """
head.listen = "tcp://localhost:4239"
head.log_file = "/dev/null"
head.gnaw.idle_quit = "10s"
head.gnaw.log_file = ""
head.gnaw.extra_args = ["-verbose"]
"""
CONFIG = load_conf(StringIO(CONFIG_STR), 'head')
async def test_exerworker(benchmark):
    async with trio.open_nursery() as _n, \
               GrainExecutor(_n=_n, nolocal=True, config=CONFIG) as exer:
        await trio.sleep(0.1) # wait for exer to start
        _n.start_soon(_worker)
        exer.submit(Cores(0), nop)
        await exer.resultq.receive() # make sure the entire pipeline is up and running
        await (asyncify(benchmark))(_parallx, exer)

async def test_remoteexer(benchmark):
    async with trio.open_nursery() as _n:
        async with RemoteExecutor(_n=_n, config=CONFIG, name="a-p") as exer:
            _n.start_soon(_worker)
            exer.submit(Cores(0), nop)
            await exer.resultq.receive() # make sure the entire pipeline is up and running
            await (asyncify(benchmark))(_parallx, exer)
            _n.cancel_scope.cancel()

def test_delayed_re(benchmark):
    async def _main():
        async def _parallx_d():
            await each([(nop_@Cores(0))() for _ in range(N)])
        nop_ = delayed(nop)
        async with trio.open_nursery() as _n:
            _n.start_soon(_worker)
            await (nop_ @ Cores(0))() # make sure the entire pipeline is up and running
            await (asyncify(benchmark))(_parallx_d)
            _n.cancel_scope.cancel()
    run(_main, config_file=StringIO(CONFIG_STR), name="a-p")


async def _worker():
    await trio.run_process([sys.executable, "-m", "grain.worker", "--url", "tcp://localhost:4239",
                            "--res", '{"Cores": { "N": 1 }}'])



import dill as pickle

from contextlib import ContextDecorator, asynccontextmanager
from timeit import default_timer as timer
from functools import wraps, partial
import types
import importlib.util
from pathlib import Path
from math import inf as INF

def timeblock(text="this block", enter=False):
    class TimeblockCtx(ContextDecorator):
        def __enter__(self):
            if enter:
                print(f"Enter {text}")
            self.st = timer()
            return self
        def __exit__(self, *exc):
            print(f"Time elapsed for {text}{' (incomplete)' if any(exc) else ''}: {timer()-self.st:.3f}s")
            return False
    return TimeblockCtx()


pickle_dumps = partial(pickle.dumps, protocol=5)
pickle_loads = pickle.loads


import trio
from trio import lowlevel as _trio
from outcome import Value

class WaitGroup(object):
    __slots__ = ("_counter", "_tasks")
    def __init__(self):
        self._counter = 0
        self._tasks = set()

    def add(self):
        self._counter += 1

    @_trio.enable_ki_protection
    def done(self, *exc):
        self._counter -= 1
        if self._counter == 0:
            for task in self._tasks:
                _trio.reschedule(task)
            self._tasks.clear()
        return False

    __enter__ = add
    __exit__ = done

    async def wait(self):
        if self._counter == 0:
            await _trio.checkpoint()
        else:
            task = _trio.current_task()
            self._tasks.add(task)
            def abort_fn(_):
                self._tasks.remove(task)
                return _trio.Abort.SUCCEEDED
            await _trio.wait_task_rescheduled(abort_fn)

PENDING = object()

class Future:
    __slots__ = ("_v", "_tasks")
    def __init__(self, v=PENDING):
        self._v = v
        self._tasks = set()

    @_trio.enable_ki_protection
    def set(self, v):
        assert self._v is PENDING
        self._v = v
        for task in self._tasks:
            _trio.reschedule(task)
        self._tasks.clear()

    async def get(self):
        if self._v is not PENDING:
            await _trio.checkpoint()
        else:
            task = _trio.current_task()
            self._tasks.add(task)
            def abort_fn(_):
                self._tasks.remove(task)
                return _trio.Abort.SUCCEEDED
            await _trio.wait_task_rescheduled(abort_fn)
        return self._v


@_trio.enable_ki_protection
def cutin_nowait(self, value):
    if self._closed:
        raise trio.ClosedResourceError
    if self._state.open_receive_channels == 0:
        raise trio.BrokenResourceError
    if self._state.receive_tasks:
        assert not self._state.data
        task, _ = self._state.receive_tasks.popitem(last=False)
        task.custom_sleep_data._tasks.remove(task)
        trio.lowlevel.reschedule(task, Value(value))
    elif len(self._state.data) < self._state.max_buffer_size:
        self._state.data.appendleft(value)
    else:
        raise trio.WouldBlock

def make_prependable(mschan):
    mschan.cutin_nowait = types.MethodType(cutin_nowait, mschan)
    return mschan


@asynccontextmanager
async def open_nursery_with_capacity(conc):
    """A patched Trio.Nursery with child task capacity
    limit. Its ``start_once_acquired`` blocks when the
    number of running child tasks started by it exceeds
    ``conc``.

    e.g.::

        async with open_nursery_with_capacity(10) as nursery:
            for _ in range(30):
                await nursery.start_once_acquired(trio.sleep, 1)
    """
    sema = trio.Semaphore(conc)
    async def _rl_task(fn, *args, task_status=trio.TASK_STATUS_IGNORED):
        async with sema:
            task_status.started()
            await fn(*args)
    async def start_once_acquired(self, fn, *args):
        await self.start(_rl_task, fn, *args)

    async with trio.open_nursery() as _n:
        _n.start_once_acquired = types.MethodType(start_once_acquired, _n)
        yield _n

@asynccontextmanager
async def open_ordered_nursery():
    """A patched Trio.Nursery that is able to start
    child task in strict first-come-first-served order,
    using its `start_now` method.
    """
    sq, rq = trio.open_memory_channel(INF)
    async def _starter():
        async for fargs in rq:
            _n.start_soon(*fargs)
    def start_now(self, *fargs):
        sq.send_nowait(fargs)

    async with trio.open_nursery() as _n:
        _n.start_soon(_starter)
        async with trio.open_nursery() as _on:
            _on.start_now = types.MethodType(start_now, _on)
            yield _on
        await sq.aclose()

@asynccontextmanager
async def QueueLimiter(conc):
    """Rate limit for the submission of delayed functions. Grain by
    default enqueues delayed functions eagerly (i.e. as soon as it
    is called). Sometimes if we have a lot of functions that can be
    run in parallel, we don't want to overwhelm the queue, so we could
    set a rate limit check prior submission. Note that the context
    scope blocks until all functions submitted through it are done.

    Args:
      conc (int): maximum number of concurrently running functions

    e.g.::

        @delayed
        async def dfn(x):
            await trio.sleep(1)
            return x+1
        r_ = 0
        async with QueueLimiter(10) as ql:
            for i in range(30):
                r_ += await ql(dfn)(i)
                #     ^^^^^wait for submission, not for result
        r = await r_
    """
    sema = trio.Semaphore(conc)
    async def sema2notify(dfn, args, kwargs, task_status):
        async with sema:
            do = dfn(*args, **kwargs)
            task_status.started(do)
            await do # sema is not released until calculation is done
    def _rl_wrapper(dfn):
        async def _rl_dfn(*args, **kwargs):
            do = await _n.start(sema2notify, dfn, args, kwargs)
            return do
        return _rl_dfn
    async with trio.open_nursery() as _n:
        yield _rl_wrapper


async def as_daemon(actx, task_status=trio.TASK_STATUS_IGNORED):
    """Transform an async context manager into a daemon task
    """
    async with actx as y:
        task_status.started(y)
        await trio.sleep_forever()


class nullacontext(object):
    def __enter__(self): return self
    def __exit__(self, *exc): return False
    async def __aenter__(self): return self
    async def __aexit__(self, *exc): return False

def optional_cm(cm, cond_arg):
    if cond_arg:
        return cm(cond_arg)
    return nullacontext()

def load_mod(modp):
    if modp == "":
        return None
    if not modp.endswith(".py"):
        return importlib.import_module(modp)
    p = Path(modp)
    spec = importlib.util.spec_from_file_location(p.stem, p)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    return mod
def load_contextmod(modp):
    if modp == "":
        return nullacontext
    return load_mod(modp).grain_context

def set_numpy_oneline_repr():
    """Change the default Numpy array ``__repr__`` to a
    compact one. This is useful for keeping the log
    clean when the job functions involve large Numpy
    arrays as args.
    """
    import numpy as np
    EDGEITEMS = 3
    def oneline_repr(a):
        N = np.prod(a.shape)
        if N == 0: # No numbers to show if the last dimension is zero
            return f"array[{a.shape}]([])"
        ind = np.unravel_index(range(min(EDGEITEMS, N)), a.shape)
        afew = []
        for x,val in zip(np.c_[ind], a[ind]):
            for i,y in enumerate(x[::-1]):
                if y != 0: break
            else: i+=1
            for j,y in enumerate(x[::-1]):
                if y+1 != a.shape[len(x)-j-1]: break
            else: j+=1
            afew.append(f"{'['*i}{val}{']'*j}")
        return f"array[{a.shape}]({', '.join(afew)}{'...' if N > EDGEITEMS else ''})"
    np.set_string_function(oneline_repr, repr=True)




N = 300 # tasks per batch

def asyncify(benchmark):
    import trio
    from functools import partial

    async def _wrapper(func, *args, **kwargs):
        await trio.to_thread.run_sync(
            benchmark,
            partial(trio.from_thread.run, partial(func, *args, **kwargs))
        )

    return _wrapper



from .delayed import *
from .resource import ZERO, Memory
from .contextvar import GVAR

import pytest
import trio
try:
    from exceptiongroup import ExceptionGroup  # for Python < 3.11
except ImportError:
    pass

from functools import partial

run = partial(run, config_file=False)

async def test_delayed_obj():
    r_ = delayedval(42)
    assert await r_ == 42
    assert await (1 + r_ * 2) == 85
    assert await (r_ + r_) == 84

    # inplace operation (setitem) creates a copy by default
    r_ = delayedval([3,4], length=2)
    a, b = r_
    r_[1] += 99
    assert (await r_ == [3,103])
    assert (await a == 3) and (await b == 4)

    # causality is respected for inplace operation
    r_ = delayedval([3,4], length=2)
    r_[1] += 99
    a, b = r_
    assert (await r_ == [3,103])
    assert (await a == 3) and (await b == 103)

async def test_numpy():
    np = pytest.importorskip('numpy')
    assert (await np.dot([2], delayedval([3]))) == 6
    assert (await np.dot(delayedval([3]), [2])) == 6
    assert (await np.dot(delayedval([2]), delayedval([3]))) == 6
    assert (await np.mean(delayedval(np.arange(10)), dtype=int)) == 4
    assert (await (np.arange(3) + delayedval(np.arange(3))).sum()) == 6
    a = np.arange(3)
    a += delayedval(np.arange(3))
    a += a
    assert (await a.sum()) == 18

    a = delayedval(np.arange(3))
    a[[0,1]] += np.array([8,9])
    assert np.allclose((await a), [8,10,2])

    a = delayedval(np.arange(3))
    a[[0,1]] += delayedval(np.array([8,9]))
    assert np.allclose((await a), [8,10,2])

    # __setitem__ on NumPy does not pass back control
    a = np.arange(3)
    with pytest.raises(TypeError):
        a[[0,1]] += delayedval(np.array([8,9]))

async def _await(r_):
    assert await r_ == 43
async def test_simutaneous_await():
    r_ = delayedval(42) + 1
    async with trio.open_nursery() as _n:
        for _ in range(3):
            _n.start_soon(_await, r_)

def test_delayed_errors():
    a = delayedval([1, 2, 3])
    # immutable
    pytest.raises(TypeError, lambda: setattr(a, "foo", 1))
    # can't iterate, or check if contains
    pytest.raises(TypeError, lambda: 1 in a)
    pytest.raises(TypeError, lambda: list(a))
    # no dynamic generation of magic/hidden methods
    pytest.raises(AttributeError, lambda: a._hidden())
    # truth of delayed forbidden
    pytest.raises(TypeError, lambda: bool(a))


@delayed
async def _anop():
    await trio.sleep(0)

class Critical(Exception):
    pass

async def _main_subtask(i):
    await _anop()
    if i == 1:
        raise Critical
def test_main_subtask_exception():
    with pytest.raises(ExceptionGroup) as excinfo:
        run([partial(_main_subtask, i) for i in range(10)], [], ZERO)
    assert excinfo.group_contains(Critical)


async def _top_serial():
    for _ in range(3):
        r = await _anop()
        assert r == None
def test_top_serial():
    run(_top_serial, [], ZERO)


@delayed
async def _a1():
    return [GVAR.res == ZERO]

async def _zero_and_more():
    assert await((_a1 @ Memory(1))() + _a1()) == [False,True]
def test_zero_and_more():
    run(_zero_and_more, [], Memory(1))


_list = []
@delayed
async def _append(i):
    _list.append(i)

async def _ordered_local_task():
    await each(reversed(list(_append(i) for i in range(10))))
    assert _list == list(range(10))
def test_ordered_local_task():
    run(_ordered_local_task, [], ZERO)



from .combine import *
from .resource import ZERO

import pytest
import trio
try:
    from exceptiongroup import ExceptionGroup  # for Python < 3.11
except ImportError:
    pass

from functools import partial

run = partial(run, config_file=False)

async def _anop():
    await trio.sleep(0)

class Critical(Exception):
    pass

async def _main_subtask(i):
    async with open_waitgroup() as wg:
        wg.submit(ZERO, _anop)
    if i == 1:
        raise Critical

def test_main_subtask_exception():
    with pytest.raises(ExceptionGroup) as excinfo:
        run([partial(_main_subtask, i) for i in range(10)], [], ZERO)
    assert excinfo.group_contains(Critical)

async def _top_serial():
    for _ in range(3):
        async with open_waitgroup() as wg:
            wg.submit(ZERO, _anop)
        assert wg.results == [None]

def test_top_serial():
    run(_top_serial, [], ZERO)



import setuptools
from distutils.command.build import build
from os import environ

exec(open("grain/_version.py").read())

class _build(build):
    def run(self):
        from subprocess import run as cmd
        cmd(
            "go mod download; "
            "CGO_ENABLED=0 GOOS=linux GOARCH=amd64 "
                f"go build -ldflags '-s -w -X main.VERSION={__version__}' -trimpath -o ../grain/gnaw",
            shell=True, check=True, cwd="gnaw/"
        )
        build.run(self)

gnaw_build_args = dict(
    cmdclass={
        'build': _build,
    },
    data_files=[
        ('bin', ['grain/gnaw',]),
    ],
) if not environ.get("CI", False) else {}

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="grain-scheduler",
    version=__version__,
    author="Harry Zhang",
    description="A scheduler for resource-aware parallel computing on clusters.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Contextualist/grain",
    packages=setuptools.find_packages(),
    python_requires=">=3.9",
    install_requires=[
        "trio >= 0.22.0",
        "dill >= 0.3.2",
        "msgpack",
        "tomli >= 1.1.0 ; python_version < '3.11'",
        "attrs >= 21.3.0",
        "cattrs",
        "click",
        "psutil",
    ],
    tests_require=[
        "pytest",
        "pytest-trio",
        "pytest-benchmark",
    ],
    entry_points={
        "console_scripts": ["grain=grain.cli:main"],
    },
    **gnaw_build_args,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Trio",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering",
        "Topic :: System :: Distributed Computing",
    ],
)



import logging
logging.getLogger('grain').setLevel(logging.DEBUG)
from .conn2 import *

import trio
import tempfile
from functools import partial

def tmpfname():
    with tempfile.NamedTemporaryFile() as f:
        return f.name

async def test_ping():
    async def _dial(edge_file):
        c = await open_tcp_edge_stream(edge_file=edge_file)
        async with c:
            await c.send_all(b"pong")
            assert (await c.receive_some()) == b"ping"
    async def _listen(edge_file, task_status=trio.TASK_STATUS_IGNORED):
        cs = trio.CancelScope()
        async def _handler(s):
            async with s:
                await s.send_all(b"ping")
                assert (await s.receive_some()) == b"pong"
            cs.cancel()
        with cs:
            await serve_tcp_edge(_handler, edge_file=edge_file, task_status=task_status)

    ef = tmpfname()
    async with trio.open_nursery() as _n:
        await _n.start(_listen, ef)
        await _dial(ef)
    async with trio.open_nursery() as _n:
        _n.start_soon(_dial, ef)
        await trio.sleep(.1)
        await _listen(ef)

async def test_double_listen(caplog):
    ef = tmpfname()
    async def _nop(s):
        pass
    async with trio.open_nursery() as _n:
        await _n.start(partial(serve_tcp_edge, _nop, edge_file=ef))
        await _n.start(partial(serve_tcp_edge, _nop, edge_file=ef))
        _n.cancel_scope.cancel()
    assert "another listener" in caplog.text



import trio
import psutil

from .conn_msgp import send_packet
from .pair import SocketChannel
from .head import GrainSpecializedRemote, backendless_sworker_manager
from .util import WaitGroup, pickle_dumps, pickle_loads, timeblock

import secrets
from functools import partial
from math import inf as INFIN
import getpass
import argparse
from pathlib import Path
import logging
logger = logging.getLogger(__name__)

DAEMON = object()

class RemoteExecutor:
    """Pass the jobs on to an external scheduler

    Args:
      gnaw (Optional[str]): If set, connect to the Gnaw executor with address
        ``gnaw``, ignoring ``head.gnaw`` in the config. By default, the Gnaw
        executor is a managed daemon process.
    """
    def __init__(self, _n=None, config=None, gnaw=DAEMON, name="", prioritized=False, sworker_config=(), **kwargs):
        if kwargs:
            logger.warning(f"kwargs {kwargs} are ignored")
        self.gnaw = gnaw
        self.name = name + ("-p" if prioritized else "")
        self.sworker_config = sworker_config
        self.push_job, self.pull_job = trio.open_memory_channel(INFIN)
        self.push_result, self.resultq = trio.open_memory_channel(INFIN)
        self.jobn = 0
        assert _n is not None
        self._n = _n
        self._c = None
        self._cs = None
        self._wg = WaitGroup() # track the entire lifetime of each job
        self.listen = config.listen
        if gnaw is DAEMON:
            self.gnaw_conf = config.gnaw
        self.side_c = None
        self.fnd = {}


    def submit(self, res, fn, *args, **kwargs):
        tid = self.jobn = self.jobn+1
        self.push_job.send_nowait((tid, res, partial(fn, *args, **kwargs)))
        return tid

    async def sealer(self):   # out of the executor scope => end of submission
        await self._wg.wait() # no pending task => end of resubmission
        await self.push_job.aclose()
        await self._c.aclose()
        await self.side_c.aclose()
    async def _sender(self):
        async with self.pull_job: # XXX: currently no ratelimit on sending
            async for tid, res, fargs in self.pull_job:
                self._wg.add()
                if not hasattr(res, "N"): # heuristic to determine if this is a sworker task
                    self.fnd[tid], func = fargs, None
                else:
                    func = pickle_dumps(fargs)
                await send_packet(self._c._so, dict(tid=tid, res=res, func=func)) # We don't need locking
    async def run(self, task_status=trio.TASK_STATUS_IGNORED):
        if self.gnaw is DAEMON: # gnaw daemon with unixconn
            # process detection
            whoami = getpass.getuser()
            for p in psutil.process_iter(['cmdline', 'username', 'name']):
                if p.info['username'] != whoami or p.info['name'] != 'gnaw':
                    continue
                pargs = p.info['cmdline']
                parser = argparse.ArgumentParser()
                parser.add_argument('-wurl', '--wurl')
                parser.add_argument('-hurl', '--hurl')
                r, _ = parser.parse_known_args(pargs)
                if r.wurl != self.listen:
                    continue
                if r.hurl is None:
                    raise RuntimeError(f"The running Gnaw instance {pargs} does not have a `hurl` arg")
                logger.info(f"found a running Gnaw instance for {self.listen!r}, going to attach")
                self.gnaw = r.hurl
                break
            else:
                logger.info(f"starting a Gnaw instance at {self.listen!r}")
                self.gnaw = f"unix:///tmp/gnaw-{secrets.token_urlsafe()}"
                conf = self.gnaw_conf
                gnawproc = await trio.lowlevel.open_process(
                    ["gnaw", "-hurl", self.gnaw, "-wurl", self.listen, "-n", str(conf.max_conn),
                             "-log", conf.log_file, "-t", conf.idle_quit, "-swarm", str(conf.swarm), *conf.extra_args],
                    start_new_session=True, # daemon
                )
                for retry in range(7): # wait for gnaw startup
                    await trio.sleep(0.1 * 2**retry)
                    if Path(self.gnaw[len('unix://'):]).exists():
                        break
                    if gnawproc.returncode is not None:
                        raise RuntimeError(f"failed to start Gnaw, exit code {gnawproc.returncode}")
                else:
                    raise RuntimeError("Gnaw took too long to start")
        with timeblock("all jobs"):
            async with SocketChannel(self.gnaw, dial=True, _n=self._n) as self._c, \
                       self.push_result:
                # handshake
                await send_packet(self._c._so, dict(cmd="chTaskResult", name=self.name))
                hsmsg = await self._c.receive()
                assert hsmsg["cmd"] == "hsSynAck"
                # self.run_sworker_clients will create chApprovalRStatus and start all required clients before reporting started
                await self._n.start(self.run_sworker_clients, hsmsg["name"])
                self._n.start_soon(self._sender)
                task_status.started()
                async for x in self._c:
                    if x['exception']: # display exception
                        tb, exp = pickle_loads(x['result'])
                        logger.error(f"Exception from task: {x['exception']}: {exp}\n{tb}")
                        continue
                    self.push_result.send_nowait((x['tid'], pickle_loads(x['result'])))
                    self._wg.done()
            if not self._c.is_clean:
                raise RuntimeError("connection to Gnaw lost unexpectedly")
    async def __aenter__(self):
        await self._n.start(self.run)
        return self
    async def __aexit__(self, *exc):
        if any(exc):
            return False
        self._n.start_soon(self.sealer)

    async def run_sworker_clients(self, dock_id, task_status=trio.TASK_STATUS_IGNORED):
        pool = dict()
        started = False

        async def _reg(name, kwargs):
            logger.info(f"client for specialized worker {name} started")
            if backendless_stype:
                backendless_stype.discard(kwargs.get("_stype", None))
            # We don't need to manage resource or connection here
            sr = GrainSpecializedRemote(name, None, None, kwargs)
            await sr.connect(_n)
            pool[name] = sr
        async def _unreg(name):
            if (sr := pool.pop(name, None)) is not None:
                await sr.aclose()

        async def _run_task(tid, res, client_name):
            ok, r = await pool[client_name].execf(tid, res, self.fnd[tid])
            await self.side_c.send(dict(tid=tid, exception="Exception" if not ok else "", result=client_name))
            if ok:
                self.push_result.send_nowait((tid, r))
                self._wg.done()
                del self.fnd[tid]
            else:
                # TODO: buffered exp
                exp, tb = r
                logger.error(f"Exception from task: {exp}\n{tb}")

        async with trio.open_nursery() as _n:
            _, backendless_stype, _backendless_sworker_n = await _n.start(backendless_sworker_manager, self.sworker_config, self.listen)
            async with SocketChannel(self.gnaw, dial=True, _n=_n) as self.side_c:
                await send_packet(self.side_c._so, dict(cmd="chApprovalRStatus", name=dock_id))
                hsmsg = await self.side_c.receive()
                assert hsmsg["cmd"] == "hsAck"
                for name, kwargs in hsmsg["obj"].items():
                    await _reg(name, kwargs)
                if not started and not backendless_stype:
                    started = True
                    task_status.started()
                # otherwise there is any pending backendless sworker, wait until registered
                async for x in self.side_c:
                    cmd = x.get("cmd", "")
                    if cmd == "":
                        tid, res, client_name = x["tid"], x["res"], x["func"]
                        _n.start_soon(_run_task, tid, res, client_name)
                    elif cmd == "pushSWorker":
                        await _reg(x["name"], x["obj"])
                        if not started and not backendless_stype:
                            started = True
                            task_status.started()
                    elif cmd == "quitSWorker":
                        await _unreg(x["name"])
                    else:
                        logger.warning(f"specialized worker client manager received unknown command {cmd}")
            for sr in pool.values():
                await sr.aclose()
            _backendless_sworker_n.cancel_scope.cancel()



"""Dask-like Delayed object, resource-centric and async/await-native
"""
from .resource import ZERO
from .contextvar import GVAR
from .util import Future

import trio

import operator
from functools import partial
from inspect import iscoroutinefunction
from copy import copy
from collections.abc import Iterable

MAX_OPS = 100


async def each(*dos):
    r"""A convinient helper function to await on a list of delayed
    objects.

    Args:
      \*dos (\*Delayed or [Iterable[Delayed],]): list of delayed
        objects

    Returns:
      list of corresponding results
    """
    if len(dos) == 1 and isinstance(dos[0], Iterable):
        dos = list(dos[0])
    return [await do for do in dos]

def delayedval(v, length=None, copy_on_setitem=True):
    """A convinient helper function that wraps any value into a
    delayed object that immediately return that value at evaluation.
    This is useful for testing and for the following edge case in NumPy::

        r_ = afn() # `afn` is a delayed function returning a len-2 array

        a = np.arange(3)
        #a[[0,1]] += r_ # ValueError; __setitem__ cannot turn `a` into a delayed object

        a = delayedval(np.arange(3))
        a[[0,1]] += r_ # Ok

    Args:
      v (Any): value to be wrapped
      length (int): see ``nout`` of :func:`delayed`
      copy_on_setitem (bool): see ``copy_on_setitem`` of :func:`delayed`

    Returns:
      a :class:`Delayed` object evaluted to ``v``
    """
    return Delayed(Future(v), length=length, copy_on_setitem=copy_on_setitem)


def delayed(fn=None, nout=None, copy_on_setitem=True, cache_hfn=None):
    """Wraps an async function into a delayed function (:class:`DelayedFn`).
    It can be used as a decorator, or around the function directly (i.e.
    ``fn = delayed(fn)``).

    Args:
      nout (int): If set, the return value of the function is assumed
        to be a iterable with length ``nout``, so that indexing and
        unpacking on the delayed object is allowed.
      copy_on_setitem (bool): The default (True) is a safe option to
        make sure previously unpacked values from a delayed objects
        are not affected by a later setitem op on that object. This
        is achieved by making a copy of the object at result evaluation,
        so this can be turned off for performance. Unlike other operations
        that can return a separate delayed object, ``setitem`` (e.g.
        ``a[1] = 1``) is an inplace mutating operatation. The following
        snippet illustrates a situation where a copy is needed::

            r_ = afn() # assume to be a delayed function returns `[3,4]`
            a_, b_ = r_
            r_[1] = 9
            assert (await r_ == [3,9])
            assert (await a_ == 3) and (await b_ == 4) # copy_on_setitem=True
            #assert (await a_ == 3) and (await b_ == 9) # copy_on_setitem=False
    """
    if fn is None:
        return partial(delayed, nout=nout, copy_on_setitem=copy_on_setitem)
    if not iscoroutinefunction(getattr(fn, "func", fn)):
        raise TypeError("delayed only wraps async functions")
    return DelayedFn(fn, nout, copy_on_setitem)

class DelayedFn:
    """A DelayedFn submits the function it wraps and returns a
    :class:`Delayed` object when called. Resources could be bound to
    the function with ``@`` operator.

    Do not construct this directly, use the :func:`@delayed<delayed>`
    decorator instead.
    """
    __slots__ = ("fn", "nout", "copy", "res")
    def __init__(self, fn, nout, copy_on_setitem):
        self.fn = fn
        self.nout = nout
        self.copy = copy_on_setitem
        self.res = ZERO
    def __matmul__(self, res):
        self.res = res
        return self
    def __call__(self, *args, **kwargs):
        # global: exer, _gn, rch
        if 'exer' not in globals():
            raise RuntimeError("Calling a delayed function is only valid inside `grain.delayed.run`")
        ft = Future()
        if self.res is not ZERO:
            tid = exer.submit(self.res, self.fn, *args, **kwargs)
            rch[tid] = ft
        else: # Currently non-leaf tasks are not guarded by the exer. We do not retry on exception.
            # We use an ordered nursery for first-come-first-serve guarentee
            _gn.start_now(partial(_run_and_set_result, ft, self.fn, *args, **kwargs))
        self.res = ZERO
        return Delayed(ft, length=self.nout, copy_on_setitem=self.copy)

async def _run_and_set_result(ft, fn, *args, **kwargs):
    r = await fn(*args, **kwargs)
    ft.set(r)


class Delayed:
    """A Delayed object represents a value to be computed by Grain.

    Do not construct this directly, this is intended to be the return
    value of a :class:`DelayedFn`. Alternatively, use :func:`delayedval`
    for wrapping a value into a Delayed object.

    A ``Delayed`` supports most python operations, each of which creates
    another ``Delayed`` representing the result:

        * Most operators (``*``, ``-``, ``+=`` (treated as ``+``), ...)
        * Item iteration, indexing, and slicing (``a[0]``)
        * Item mutation __setitem__ (``a[0] = 1``)
        * Attribute access (``a.size``)
        * Method calls (``a.index(0)``)
        * Most NumPy functions (``np.mean(a)``)

    Operations that aren’t supported include:

        * Attr mutation __setattr__ (``a.foo = 1``)
        * Use as a predicate (``if a: ...``)

    This is mostly a copy of the implementaton in Dask, but there are
    several main differences from Dask:

        1. Grain's Delayed assumes all ops are cheap and perform them
           locally.
        2. As all delayed objects are reduced locally, and Grain worker
           does not cache results, it is not allowed to pass delayed
           objects to a delayed function; we want to be explicit on
           the intention of dependent/serial jobs. e.g.::

               r1 = await dfn1()
               r2 = await dfn2(r1)

        3. Calling a delayed function submit the calculation immediately.
        4. A Delayed object is somehow mutable, ``setitem`` is allowed
           but ``setattr`` is not (implemented).
    """
    __slots__ = ("_future", "_post_ops", "_length", "_copy", "_eval_started")
    def __init__(self, future, post_ops=None, length=None, copy_on_setitem=True):
        self._future = future
        self._post_ops = post_ops or []
        self._length = length
        self._copy = copy_on_setitem
        self._eval_started = False
        if len(self._post_ops) > MAX_OPS:
            self._partition_ops()
    def __getstate__(self):
        return tuple(getattr(self, i) for i in self.__slots__)
    def __setstate__(self, state):
        for k, v in zip(self.__slots__, state):
            setattr(self, k, v)

    async def result(self):
        if self._eval_started: # the first call does the eval; the rest wait for its result
            return await self._future.get()
        self._eval_started = True
        _base_future, self._future = self._future, Future()
        r = await _base_future.get()
        for op, other, kwargs in self._post_ops:
            #print("post_op", op, "on", other)
            for i, o in enumerate(other):
                if not isinstance(o, Delayed): continue
                other[i] = await o.result() # eval o if not done
            for k, o in kwargs.items():
                if not isinstance(o, Delayed): continue
                kwargs[k] = await o.result() # eval o if not done
            if op is not operator.setitem: # NOTE: exclude the inplace op
                r = op(r, *other, **kwargs)
            else:
                if self._copy:
                    r = copy(r) # shallow copy is ok because setitem does not change the value itself
                op(r, *other)
        self._future.set(r)
        self._post_ops.clear()
        return r
    def __await__(self):
        """Await on the Delayed object returns its result"""
        return self.result().__await__()
    get = result # implements the Future interface

    def _partition_ops(self):
        chunk_ops, self._post_ops = self._post_ops[:MAX_OPS], self._post_ops[MAX_OPS:]
        self._future = Delayed(self._future, chunk_ops, self._length, self._copy)
        _gn.start_soon(self._future.get) # kickstart the chunked eval

    def __dir__(self):
        return dir(type(self))

    def __getattr__(self, attr):
        if attr.startswith("_"):
            raise AttributeError(f"Attribute {attr} not found")
        return self._get_operator(getattr)(self, attr)

    def __setattr__(self, attr, val):
        if attr in self.__slots__:
            object.__setattr__(self, attr, val)
        else:
            raise TypeError("setattr on Delayed objects is not allowed")

    def __setitem__(self, index, val):
        #print("memorize op", operator.setitem, [index,val])
        assert not self._eval_started, "cannot add inplace operation __setitem__ after evaluation has started"
        self._post_ops.append((operator.setitem,[index,val],{})) # setitem is an inplace op
        if len(self._post_ops) >= MAX_OPS:
            self._partition_ops()

    def __iter__(self):
        if self._length is None:
            raise TypeError("Delayed objects of unspecified length are not iterable")
        for i in range(self._length):
            yield self[i]

    def __len__(self):
        if self._length is None:
            raise TypeError("Delayed objects of unspecified length have no len()")
        return self._length

    def __bool__(self):
        raise TypeError("Truth of Delayed objects is not supported")
    __nonzero__ = __bool__

    def __get__(self, instance, cls):
        if instance is None:
            return self
        return types.MethodType(self, instance)

    def __array_function__(self, func, _types, args, kwargs): # NumPy function support
        return self._get_operator(_deself(func))(self, *args, **kwargs)
    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs): # NumPy ufunc support
        # Simply setting the __array_ufunc__ method to None does not support inplace op (e.g. +=)
        return self._get_operator(_deself(getattr(ufunc, method)))(self, *inputs, **kwargs)

    @classmethod
    def _get_operator(cls, op):
        """Returns the memorized version of op
        """
        def mem_op(self, *other, **kwargs): # record the op in the instance returned
            #print("memorize op", op, other)
            return cls(self._future, [*self._post_ops, (op,list(other),kwargs)], self._length, self._copy)
        return mem_op

def _deself(op):
    """For NumPy function / ufunc, self is amid the args, and the first self is redundant"""
    def _inner(_self, *args, **kwargs):
        return op(*args, **kwargs)
    return _inner

Delayed.__call__ = Delayed._get_operator(lambda fn, *args, **kwargs: fn(*args, **kwargs))

def _right(op):
    """Wrapper to create 'right' version of operator given left version"""
    def _inner(self, other):
        return op(other, self)
    return _inner

def _bind_operator(cls, op):
    """Bind operator to class"""
    name = op.__name__
    if name.endswith("_"):
        # for and_ and or_
        name = name[:-1]
    elif name == "inv":
        name = "invert"
    meth = f"__{name}__"
    setattr(cls, meth, cls._get_operator(op))
    if name not in (
        'add', 'and', 'divmod', 'floordiv', 'lshift', 'matmul', 'mod',
        'mul', 'or', 'pow', 'rshift', 'sub', 'truediv', 'xor',
    ):
        return
    rmeth = f"__r{name}__"
    setattr(cls, rmeth, cls._get_operator(_right(op)))
for op in (
    operator.abs,
    operator.neg,
    operator.pos,
    operator.invert,
    operator.add,
    operator.sub,
    operator.mul,
    operator.floordiv,
    operator.truediv,
    operator.mod,
    operator.pow,
    operator.and_,
    operator.or_,
    operator.xor,
    operator.lshift,
    operator.rshift,
    operator.eq,
    operator.ge,
    operator.gt,
    operator.ne,
    operator.le,
    operator.lt,
    operator.getitem,
    operator.matmul,
):
    _bind_operator(Delayed, op)


async def relay(inq):
    async with inq:
        async for i, rslt in inq:
            rch[i].set(rslt)
            del rch[i] # one job, one future
async def boot(subtasks, args, kwargs):
    global exer, _gn, rch
    from .head import GrainExecutor
    from .remote_exer import RemoteExecutor
    from .config import load_conf, load_conf_sworker
    from .delayed import relay # no global dependency
    from .util import open_ordered_nursery
    from collections.abc import Iterable
    if (local := kwargs.pop('local', None)) is not None:
        kwargs['rpw'] = local
    config_file = kwargs.pop('config_file', None)
    config = load_conf(config_file, 'head')
    kwargs['sworker_config'] = load_conf_sworker(config_file) if config_file is not False else []
    Exer = RemoteExecutor
    if local or not config.gnaw.enabled:
        Exer = GrainExecutor
    async with trio.open_nursery() as _n, \
               Exer(_n=_n, config=config, *args, **kwargs) as exer:
        rch = {}
        _n.start_soon(relay, exer.resultq)
        GVAR.instance = "N/A"
        GVAR.res = ZERO
        async with open_ordered_nursery() as _gn:
            if isinstance(subtasks, Iterable):
                for st in subtasks:
                    _gn.start_soon(st)
            else:
                await subtasks()

def run(subtasks, *args, **kwargs):
    """Delayed's main entry point. Start your "root" async function(s) here.
    Except for the first two args, all the other args are optional and are passed
    to :class:`grain.remote_exer.RemoteExecutor` (or to :class:`grain.head.GrainExecutor`
    if ``local`` is set or Gnaw is disabled in the config).
    Common options are listed below, for more options (usually for internal debug),
    see the executors.

    Args:
      subtasks: Root async function(s) that spawns all the other calculations.
        This could be an async function or an iterable of async functions. If
        an iterable is passed, all functions are run concurrently.
      local (Optional[~grain.resource.Resource]): Resource for local worker.
        Default to None (Gnaw executor has no local worker). If set, the built-in
        :class:`grain.head.GrainExecutor` will be used.
      config_file (str | Literal[False] | None): Grain's config file name. If
        not set or None, Grain will use the name provided by envar ``GRAIN_CONFIG``,
        and finally fallback to name ``grain.toml``. If set to False, Grain will
        use the default profile (see ``config.py``).
      name (str): For Gnaw executor only. Name to be recognized in the Gnaw's log
      prioritized (bool): Default to False. For Gnaw executor only. If true,
        submit tasks to the prioritized queue.
    """
    trio.run(boot, subtasks, args, kwargs)



from grain.head import GrainExecutor
from grain.config import load_conf
from grain.resource import Memory, ZERO
import trio

import pytest
from . import asyncify, N

async def nop():
    await trio.sleep(0)

async def _parallx(exer):
    for _ in range(N):
        exer.submit(Memory(0), nop)
    for _ in range(N):
        await exer.resultq.receive()

async def test_head(benchmark):
    async with trio.open_nursery() as _n, \
               GrainExecutor(_n=_n, rpw=Memory(0), config=load_conf(False,'head')) as exer:
        await (asyncify(benchmark))(_parallx, exer)



from .resource import *

import pytest

def test_zero():
    # repr
    assert str(ZERO) == "Ø"
    # request, alloc
    assert ZERO.request(ZERO)
    assert str(ZERO.alloc(ZERO)) == "Ø"
    some = Node(1,1)
    assert some.request(ZERO)
    assert str(some.alloc(ZERO)) == "Ø"
    assert ZERO.request(some) is False
    with pytest.raises(ValueError, match="Cannot alloc"):
        ZERO.alloc(some)

def test_cores():
    # repr
    assert str(Cores(3)) == "CPU_Cores([0-2])"
    assert str(Cores([5,7,8])) == "CPU_Cores([5,7-8])"
    # request, alloc
    c0 = Cores(16)
    assert c0.request(Cores(16))
    assert c0.request(Cores(17)) is False
    with pytest.raises(ValueError, match=r"cannot allocate .* core\(s\)"):
        c0.alloc(Cores(17))
    a = c0.alloc(Cores(3))
    assert len(a.c)==3 and a.c&c0.c==set()
    assert c0.request(Cores(14)) is False
    # dealloc
    c0.dealloc(a)
    assert len(c0.c)==c0.N==16

def test_memory():
    # repr
    assert str(Memory(16)) == "Memory(16GB)"
    # request, alloc
    m0 = Memory(64)
    assert m0.request(Memory(64))
    assert m0.request(Memory(65)) is False
    with pytest.raises(ValueError, match="cannot allocate"):
        m0.alloc(Memory(65))
    a = m0.alloc(Memory(16))
    assert a.m==16 and a.m+m0.m==64
    assert m0.request(Memory(50)) is False
    # dealloc
    m0.dealloc(a)
    assert m0.m==m0.M==64

def test_walltime(monkeypatch):
    import time
    monkeypatch.setattr(time, "time", lambda: now) # mock timestamp
    # repr
    now = 0.0
    t0, t1, t2 = WTime(T="1:56:40"), WTime(T=7000), WTime(T=7000, countdown=True)
    now += 2.0
    assert str(t0)==str(t1)=="Walltime(01:56:40)" and str(t2)=="Walltime(01:56:38)"
    # request, alloc
    now = 0.0
    t2 = WTime(T=30, countdown=True)
    assert t2.request(WTime(T=30))
    now += 2.0
    assert t2.request(WTime(T=29)) is False
    assert t2.alloc(WTime(T=3)).T == 3
    # group
    from .resource import TIMESTAT_NSAMPLE
    now = 0.0
    t3, t4 = WTime(T=7000, countdown=True), WTime(group='tag')
    a = [t3.alloc(t4) for _ in range(TIMESTAT_NSAMPLE)]
    now += 12.0
    [t3.dealloc(x) for x in a]
    infr = t3.alloc(t4)
    assert infr.T == infr.softT == 12.0

def test_token():
    # repr
    assert str(Token('A')) == "Token(A)"
    # request, alloc
    a = Token('A')
    assert a.request(Token('A'))
    assert a.request(Token('B')) is False
    assert a.alloc(Token('A')) == a
    assert a.request(Token('A'))

def test_capacity():
    # repr
    assert str(Capacity(0)) == "Capacity(0)"
    assert str(ONE_INSTANCE) == "ONE"
    # request, alloc
    c = Capacity(1)
    assert c.request(ONE_INSTANCE)
    assert c.alloc(ONE_INSTANCE).v == -1
    assert c.v==0
    with pytest.raises(ValueError, match="has no capacity"):
        c.alloc(ONE_INSTANCE)
    # dealloc
    c.dealloc(ONE_INSTANCE)
    assert c.v==c.V==1

def test_reject():
    # request
    assert REJECT.request(ZERO) is False
    r = Reject(Memory(8))
    r.dealloc(Memory(4))
    assert r.request(Memory(8)) is False
    # eject
    assert r.eject() == Memory(12)

def test_multiresource():
    # repr
    assert str(Cores(3) & Memory(16)) == "CPU_Cores([0-2]) & Memory(16GB)"
    # request, alloc
    r = Cores(6) & Memory(16)
    assert r.request(Memory(8))
    assert r.request(Cores(3) & Memory(8))
    assert r.request(Cores(9)) is False
    assert r.request(Cores(3) & Memory(17)) is False
    with pytest.raises(ValueError, match="cannot allocate"):
        r.alloc(Cores(3) & Memory(17))
    with pytest.raises(ValueError, match="does not have the same shape"):
        r.alloc(WTime(T=3))
    a = r.alloc(Cores(3) & Memory(8))
    assert len(a.c)==3 and a.m==8 and len(a.c|r.c)==6 and a.m+r.m==16
    assert r.request(Cores(3) & Memory(9)) is False
    # dealloc
    r.dealloc(a)
    assert len(r.c)==r.N==6 and r.m==r.M==16
